from __future__ import annotations

import json
from pathlib import Path
from typing import TypedDict, Literal

from neuralib.util.util_type import PathLike

__all__ = [
    'MonitorDict',
    'MinotorWarpDict',
    'FlashIndicatorParameters',
    'NetworkControllerDict',
    'RigDict',
    'PreferenceDict',
    #
    'load_preferences'
]


class MonitorDict(TypedDict, total=False):
    sizePix: list[int]
    fullScreen: bool

    # PsychoPyDisplay
    name: str
    rate: float
    distance: int
    gamma: float
    width: int
    screen: int
    pos: list[int]
    winType: str

    # PyGameVRDisplay
    fov: float
    n_rays: float
    max_depth: float


class MinotorWarpDict(TypedDict, total=False):
    eyepoint: list[float]
    flipHorizontal: bool
    flipVertical: bool
    warp: str
    warpGridsize: int
    warpfile: str


class FlashIndicatorParameters(TypedDict, total=False):
    size: int
    units: str
    pos: list[int]
    fillColor: float
    mode: int
    state: bool
    frames: int


class NetworkControllerDict(TypedDict, total=False):
    ip: str
    port: int
    trigger: bool


class RigDict(TypedDict):
    port: str | Literal['dummy']


class PreferenceDict(TypedDict, total=False):
    user: str
    userPrefix: str
    expname: str  # git only
    defaultExperimentType: str
    default_imaging_mode: str

    logFolder: str | Path
    protocolsFolder: str | Path
    controllerFolder: str | Path
    stimsFolder: str | Path
    tmpFolder: str | Path  # git only

    monitor: list[MonitorDict]
    use_monitor: int | list[int]

    # PanoDisplay
    vr_flag: bool
    vrFolder: str | Path

    # PyGameVRDisplay
    textureFolder: str | Path
    raycaster: Literal['default', 'numpy', 'numba']

    # Photo indicator
    flashIndicator: bool
    flashIndicatorMode: int
    flashIndicatorParameters: FlashIndicatorParameters

    # Network
    labcams: NetworkControllerDict
    scanbox: NetworkControllerDict
    pycams: NetworkControllerDict
    spikeglx: NetworkControllerDict

    rig: RigDict
    warp: MinotorWarpDict

    # runtime append
    _controllers_data_folder: Path
    _controllers_protocol_folder: str
    _controllers_folder_read_flag: bool


def load_preferences(file: PathLike) -> PreferenceDict:
    """
    Get stimpy output(parsed) preference file as dict

    :param file: filepath for the .pref
    :return: :class:`PreferenceDict`
    """
    if Path(file).suffix != '.pref':
        raise ValueError('invalid file type')

    with open(file, "r") as file:
        return json.load(file)



from __future__ import annotations

from pathlib import Path
from typing import Literal, NamedTuple, Iterable

import numpy as np
import pandas as pd
import polars as pl

from neuralib.atlas.map import NUM_MERGE_LAYER
from neuralib.atlas.type import Source, HEMISPHERE_TYPE
from neuralib.util.util_type import DataFrame
from neuralib.util.util_verbose import fprint

__all__ = [
    'ALLEN_CCF_10um_BREGMA',
    'ALLEN_SOURCE_TYPE',
    'PLANE_TYPE',
    #
    'SourceCoordinates',
    'iter_source_coordinates',
    'get_margin_merge_level',
    'roi_points_converter',
]

# allen CCF 10um volume coordinates, refer to allenCCF/Browsing Functions/allenCCFbregma.m
ALLEN_CCF_10um_BREGMA = np.array([540, 0, 570])  # AP, DV, LR
ALLEN_SOURCE_TYPE = Literal['annotation', 'nrrd', 'template']

PLANE_TYPE = Literal['coronal', 'sagittal', 'transverse']


class SourceCoordinates(NamedTuple):
    source: Source
    coordinates: np.ndarray
    """(N, 3) with ap, dv, ml"""

    axes_repr: tuple[str, str, str] = ('ap', 'dv', 'ml')

    @property
    def ap(self) -> np.ndarray:
        return self.coordinates[:, 0]

    @property
    def dv(self) -> np.ndarray:
        return self.coordinates[:, 1]

    @property
    def ml(self) -> np.ndarray:
        return self.coordinates[:, 2]


def iter_source_coordinates(
        file: Path,
        *,
        only_areas: list[str] | str | None = None,
        region_col: str | None = None,
        hemisphere: HEMISPHERE_TYPE = 'both',
        to_brainrender: bool = True,
        to_um: bool = True,
        ret_order: tuple[Source, ...] | None = ('pRSC', 'aRSC', 'overlap')
) -> Iterable[SourceCoordinates]:
    """Load allen ccf roi output (merged different color channels).

    :param file: parsed csv file after
    :param only_areas: only show rois in region(s)
    :param region_col: if None, auto infer, and check the lowest merge level contain all the regions specified
    :param hemisphere
    :param to_brainrender: convert the coordinates to brain render
    :param to_um
    :param ret_order: whether specify the source generator order
    :return :class:`SourceCoordinates`
    """
    df = pl.read_csv(file)
    #
    if only_areas is not None and len(only_areas) != 0:
        if isinstance(only_areas, str):
            only_areas = [only_areas]

        if region_col is None:
            region_col = get_margin_merge_level(df, only_areas, 'lowest')

        df = df.filter(pl.col(region_col).is_in(only_areas))
        fprint(f'using {region_col} for {only_areas}')

        if df.is_empty():
            raise RuntimeError('check lowest merge level')
    #
    if hemisphere != 'both':
        df = df.filter(pl.col('hemi.') == hemisphere)

    #
    points = roi_points_converter(df, to_brainrender=to_brainrender, to_um=to_um)

    if ret_order is None:
        ret_order = df['source'].unique()

    for src in ret_order:
        mask = df.select(pl.col('source') == src).to_numpy()[:, 0]
        yield SourceCoordinates(src, points[mask])


def get_margin_merge_level(df: pl.DataFrame,
                           areas: list[str] | str,
                           margin: Literal['lowest', 'highest']) -> str:
    """Get the lowest or highest merge level (i.e., parsed_csv) containing all the regions

    :param df: parsed csv
    :param areas: an area or a list of areas
    :param margin: get the either lowest of highest merge level for a given area
    :return col name if parsed csv
    """
    if not isinstance(areas, (tuple, list)):
        areas = [areas]

    eval_merge = [f'merge_ac_{i}' for i in range(NUM_MERGE_LAYER)]

    if margin == 'lowest':
        level = eval_merge
    elif margin == 'highest':
        level = reversed(eval_merge)
    else:
        raise ValueError('')

    for lv in level:
        ls = df[lv]
        if np.all([a in ls for a in areas]):
            return lv

    raise ValueError(f'{areas} not found')


def roi_points_converter(dat: DataFrame | np.ndarray,
                         to_brainrender: bool = True,
                         to_um: bool = True) -> np.ndarray:
    """
    convert coordinates of `allenccf` roi points from parsed dataframe

    :param dat: Dataframe with 'AP_location', 'DV_location', 'ML_location' headers.
            Or numpy array with (N, 3), or (3,)
    :param to_brainrender: coordinates to `brainrender`
    :param to_um:

    :return (N, 3),
        N: number of roi
        3: ap, dv, ml
    """
    if isinstance(dat, pd.DataFrame):
        dat = pl.from_pandas(dat)
        points = dat.select('AP_location', 'DV_location', 'ML_location').to_numpy()
    elif isinstance(dat, pl.DataFrame):
        points = dat.select('AP_location', 'DV_location', 'ML_location').to_numpy()
    elif isinstance(dat, np.ndarray):
        if dat.ndim == 1:
            dat = np.expand_dims(dat, axis=0)

        if dat.ndim == 2 and dat.shape[1] != 3:
            raise ValueError('')

        points = dat
    else:
        raise TypeError('')

    #
    if to_um:
        points *= 1000  # um

    if to_brainrender:
        points[:, 0] /= -1  # increment toward posterior
        points[:, 2] /= -1  # increment toward left hemisphere
        bregma = ALLEN_CCF_10um_BREGMA * 10  # pixel to um
        points += bregma  # roi relative to bregma

    return points



import unittest
from pathlib import Path
from typing import Optional, Literal

from neuralib.argp import *


class TypeAnnotationTest(unittest.TestCase):
    def test_bool(self):
        class Opt:
            a: bool = argument('-a')

        opt = parse_args(Opt(), ['-a'])
        self.assertTrue(opt.a)

        opt = parse_args(Opt(), [])
        self.assertFalse(opt.a)

    def test_bool_set_false(self):
        class Opt:
            a: bool = argument('-a', default=True)

        opt = parse_args(Opt(), ['-a'])
        self.assertFalse(opt.a)

        opt = parse_args(Opt(), [])
        self.assertTrue(opt.a)

    def test_str(self):
        class Opt:
            a: str = argument('-a')

        opt = parse_args(Opt(), ['-a', 'test'])
        self.assertEqual(opt.a, 'test')
        opt = parse_args(Opt(), [])
        self.assertIsNone(opt.a)

    def test_optional_str(self):
        class Opt:
            a: Optional[str] = argument('-a')

        opt = parse_args(Opt(), ['-a', 'test'])
        self.assertEqual(opt.a, 'test')
        opt = parse_args(Opt(), [])
        self.assertIsNone(opt.a)

    def test_int(self):
        class Opt:
            a: int = argument('-a')

        opt = parse_args(Opt(), ['-a', '10'])
        self.assertEqual(opt.a, 10)
        opt = parse_args(Opt(), [])
        self.assertIsNone(opt.a)

    def test_float(self):
        class Opt:
            a: float = argument('-a')

        opt = parse_args(Opt(), ['-a', '10.321'])
        self.assertEqual(opt.a, 10.321)
        opt = parse_args(Opt(), [])
        self.assertIsNone(opt.a)

    def test_path(self):
        class Opt:
            a: Path = argument('-a')

        opt = parse_args(Opt(), ['-a', 'test_argp.py'])
        self.assertEqual(opt.a, Path('test_argp.py'))
        opt = parse_args(Opt(), [])
        self.assertIsNone(opt.a)

    def test_literal(self):
        class Opt:
            a: Literal['A', 'B'] = argument('-a')

        opt = parse_args(Opt(), ['-a', 'A'])
        self.assertEqual(opt.a, 'A')

        opt = parse_args(Opt(), [])
        self.assertIsNone(opt.a)

        with self.assertRaises(SystemExit):
            parse_args(Opt(), ['-a', 'C'])

    def test_optional_literal(self):
        class Opt:
            a: Optional[Literal['A', 'B']] = argument('-a')

        opt = parse_args(Opt(), ['-a', 'A'])
        self.assertEqual(opt.a, 'A')

        opt = parse_args(Opt(), [])
        self.assertIsNone(opt.a)

        with self.assertRaises(SystemExit):
            parse_args(Opt(), ['-a', 'C'])

    def test_literal_with_choice(self):
        class Opt:
            a: Literal['A', 'B'] = argument('-a', choices=('A', 'B', 'C'))

        opt = parse_args(Opt(), ['-a', 'C'])
        self.assertEqual(opt.a, 'C')

    def test_list_type_extend(self):
        class Opt:
            a: list[str] = argument(metavar='...', nargs='*', action='extend')

        opt = parse_args(Opt(), [])
        self.assertListEqual(opt.a, [])

        opt = parse_args(Opt(), ['12', '34'])
        self.assertListEqual(opt.a, ['12', '34'])

    def test_list_type_append(self):
        class Opt:
            a: list[str] = argument('-a', action='append')

        opt = parse_args(Opt(), [])
        # self.assertListEqual(opt.a, [])
        self.assertIsNone(opt.a) # should it be a []

        opt = parse_args(Opt(), ['-a=1'])
        self.assertListEqual(opt.a, ['1'])

        opt = parse_args(Opt(), ['-a=1', '-a=2'])
        self.assertListEqual(opt.a, ['1', '2'])

    def test_list_type_infer(self):
        class Opt:
            a: list[int] = argument(metavar='...', nargs='*', action='extend')

        opt = parse_args(Opt(), ['12', '34'])
        self.assertListEqual(opt.a, [12, 34])


class WithDefaultTest(unittest.TestCase):
    def test_bool(self):
        class Opt:
            a: bool = argument('-a')

        opt = with_defaults(Opt())
        self.assertFalse(opt.a)

    def test_bool_set_false(self):
        class Opt:
            a: bool = argument('-a', default=True)

        opt = with_defaults(Opt())
        self.assertTrue(opt.a)

    def test_str(self):
        class Opt:
            a: str = argument('-a')

        opt = with_defaults(Opt())
        self.assertIsNone(opt.a)

    def test_default_str(self):
        class Opt:
            a: str = argument('-a', default='default')

        opt = with_defaults(Opt())
        self.assertEqual(opt.a, 'default')

    def test_int(self):
        class Opt:
            a: int = argument('-a')

        opt = with_defaults(Opt())
        self.assertIsNone(opt.a)

    def test_default_int(self):
        class Opt:
            a: int = argument('-a', default=101)

        opt = with_defaults(Opt())
        self.assertEqual(opt.a, 101)

    def test_float(self):
        class Opt:
            a: float = argument('-a')

        opt = with_defaults(Opt())
        self.assertIsNone(opt.a)

    def test_default_float(self):
        class Opt:
            a: float = argument('-a', default=3.14)

        opt = with_defaults(Opt())
        self.assertEqual(opt.a, 3.14)

    def test_literal(self):
        class Opt:
            a: Literal['A', 'B'] = argument('-a')

        opt = with_defaults(Opt())
        self.assertIsNone(opt.a)

    def test_default_literal(self):
        class Opt:
            a: Literal['A', 'B'] = argument('-a', default='C')

        opt = with_defaults(Opt())
        self.assertEqual(opt.a, 'C')


class AsDictTest(unittest.TestCase):
    def test_emtpy(self):
        class Opt:
            a: str = argument('-a', default='default')

        self.assertDictEqual(as_dict(Opt()), {})

    def test_as_dict(self):
        class Opt:
            a: str = argument('-a', default='default')

        opt = with_defaults(Opt())
        self.assertDictEqual(as_dict(opt), {'a': 'default'})


class AbstractParserTest(unittest.TestCase):
    def test_exit_on_error(self):
        class Main(AbstractParser):
            a: str = argument('-a', default='default')

        with self.assertRaises(SystemExit):
            Main().main(['-b'], exit_on_error=True)

        with self.assertRaises(RuntimeError):
            Main().main(['-b'], exit_on_error=False)


class CommandParserTest(unittest.TestCase):
    def test_command_parser(self):
        class P1(AbstractParser):
            a: str = argument('-a', default='default')

        class P2(AbstractParser):
            a: str = argument('-a', default='default')

        parsers = dict(a=P1, b=P2)
        opt = parse_command_args(parsers, ['a'], run_main=False)
        self.assertIsInstance(opt, P1)
        opt = parse_command_args(parsers, ['b'], run_main=False)
        self.assertIsInstance(opt, P2)
        opt = parse_command_args(parsers, [], run_main=False)
        self.assertIsNone(opt)

    def test_command_parser_main(self):
        class P1(AbstractParser):
            a: str = argument('-a', default='default')

        class P2(AbstractParser):
            a: str = argument('-a', default='default')

        parsers = dict(a=P1, b=P2)
        opt = parse_command_args(parsers, ['a'])
        self.assertIsInstance(opt, P1)
        self.assertEqual('default', opt.a)
        opt = parse_command_args(parsers, ['b'])
        self.assertIsInstance(opt, P2)
        self.assertEqual('default', opt.a)
        opt = parse_command_args(parsers, [])
        self.assertIsNone(opt)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import dataclasses
import re
from pathlib import Path
from typing import final

import attrs
import numpy as np
import polars as pl
from scipy.interpolate import interp1d

from neuralib.util.util_type import PathLike
from neuralib.util.util_verbose import fprint
from .baselog import Baselog, LOG_SUFFIX, StimlogBase, AbstractStimTimeProfile
from .baseprot import AbstractStimProtocol
from .session import Session, SessionInfo
from .stimulus import StimPattern
from .util import try_casting_number, unfold_stimuli_condition

__all__ = ['PyVlog',
           'StimlogPyVStim',
           'PyVProtocol']


@final
class PyVlog(Baselog):
    """class for handle the log file (rig event specific) for pyvstim version (vb lab legacy)"""

    def __init__(self,
                 root_path: PathLike,
                 log_suffix: LOG_SUFFIX = '.log',
                 diode_offset: bool = False):

        super().__init__(root_path, log_suffix, diode_offset)
        self.__prot_cache: PyVProtocol | None = None

    @classmethod
    def _cache_asarray(cls, filepath: Path) -> np.ndarray:
        output = filepath.with_name(filepath.stem + '_log.npy')

        if not output.exists():

            data_list = []
            with filepath.open() as f:
                for line, content in enumerate(f):
                    content = content.strip()
                    if not content.startswith('#') and content != '':  # comments and empty line
                        cols = content.strip().split(',')
                        # Convert the columns to floats
                        cols = [float(x) for x in cols]
                        # Append the row to data_list
                        data_list.append(cols)

            # Find the maximum number of columns
            max_cols = max([len(row) for row in data_list])

            new_data = []

            # Iterate over each row
            for row in data_list:
                # Calculate the number of columns to add
                cols_to_add = max_cols - len(row)
                # Add the required number of np.nan values
                row.extend([np.nan] * cols_to_add)
                # Append the row to new_data
                new_data.append(row)

            # Convert new_data to a numpy array
            ret = np.array(new_data)

            np.save(output, ret)

        return np.load(output)

    # ===== #

    def stimlog_data(self) -> 'StimlogPyVStim':
        return StimlogPyVStim(self)

    def get_prot_file(self) -> PyVProtocol:
        if self.__prot_cache is None:
            self.__prot_cache = PyVProtocol.load(self.stim_prot_file)

        return self.__prot_cache


# ======= #
# Stimlog #
# ======= #

@final
class StimlogPyVStim(StimlogBase):
    """class for handle the log file (stim event specific) for pyvstim version (vb lab legacy)

    `Dimension parameters`:

        N = number of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = number of Stim Type

        P = number of acquisition sample pulse

    """

    v_present_time: np.ndarray
    """(P,) in ms"""
    v_stim: np.ndarray
    """(P,), stim type index. value from 1 to S"""
    v_trial: np.ndarray
    """(P,) number of trial. value from 1 to R. *last is 1, reset?"""
    v_frame: np.ndarray
    """(P,), value 0, 1, 2... TBD"""
    v_blank: np.ndarray
    """(P,). whether is background only. 0: stim display, 1: no stim"""
    v_contrast: np.ndarray
    """(P, ) background contrast, 0 to 1?"""
    v_pos_x: np.ndarray
    """(P, ) display pos x"""
    v_pos_y: np.ndarray
    """(P, ) display pos y"""
    v_ap_x: np.ndarray
    """(P, ) stim center x"""
    v_ap_y: np.ndarray
    """(P, ) stim center y"""
    v_indicator_flag: np.ndarray
    """(P,). photo-indicator, for stim-onset. 1: stim display, 0: no stim"""
    v_duino_time: np.ndarray
    """(P,). extrapolate duinotime from screen indicator. sync arduino time in sec"""

    def __init__(self, riglog: 'PyVlog'):
        super().__init__(riglog, file_path=None)
        self._reset()

    def _reset(self):
        _attrs = (
            'v_present_time',
            'v_stim',
            'v_trial',
            'v_frame',
            'v_blank',
            'v_contrast',
            'v_pos_x',
            'v_pos_y',
            'v_ap_x',
            'v_ap_y',
            'v_indicator_flag'
        )
        code = self.riglog_data.dat[:, 0] == 10
        for i, it in enumerate(self.riglog_data.dat[code, 1:].T):
            setattr(self, f'{_attrs[i]}', it)

        self.v_duino_time = self._get_stim_duino_time(self.riglog_data.dat[code, -1].T)

    def _get_stim_duino_time(self, indicator_flag: np.ndarray) -> np.ndarray:
        """extrapolate duinotime from screen indicator. sync arduino time in (P,) sec"""
        fliploc = np.where(
            np.diff(np.hstack([0, indicator_flag, 0])) != 0
        )[0]

        return interp1d(
            fliploc,
            self.riglog_data.screen_event.time,
            fill_value="extrapolate"
        )(np.arange(len(indicator_flag)))

    @property
    def exp_start_time(self) -> float:
        return self.v_duino_time[0]

    @property
    def stimulus_segment(self) -> np.ndarray:
        return self.get_time_profile().get_time_interval()

    def session_trials(self) -> dict[Session, SessionInfo]:
        raise NotImplementedError('')

    @property
    def time_offset(self) -> float:
        """directly used interpolation using diode signal already"""
        return 0

    def get_stim_pattern(self) -> StimPattern:
        raise NotImplementedError('')

    def exp_end_time(self) -> float:
        return self.v_duino_time[-1]

    # =========== #
    # retinotopic #
    # =========== #

    @property
    def stim_loc(self) -> np.ndarray:
        return np.vstack([self.v_ap_x, self.v_ap_y]).T

    @property
    def avg_refresh_rate(self) -> float:
        """in Hz"""
        return 1 / (np.diff(self.v_duino_time).mean())

    def plot_stim_animation(self):
        from neuralib.plot.animation import plot_scatter_animation
        plot_scatter_animation(self.v_ap_x,
                               self.v_ap_y,
                               self.v_duino_time,
                               step=int(self.avg_refresh_rate))  # TODO check refresh rate?

    def get_time_profile(self) -> StimTimeProfile:
        return StimTimeProfile(self)


@final
@attrs.frozen(repr=False, str=False)
class StimTimeProfile(AbstractStimTimeProfile):
    stim: StimlogPyVStim

    def __repr__(self):
        ret = {
            'n_trials': self.n_trials,
            'foreach_trial_time': np.diff(self.get_time_interval(), axis=1),
            'n_cycle:': self.stim.riglog_data.get_prot_file().get_loops_expr().n_cycles
        }
        return repr(ret)

    __str__ = __repr__

    @property
    def unique_stimuli_set(self) -> pl.DataFrame:
        df = pl.DataFrame({
            'i_stims': self.stim.v_stim.astype(int),
            'i_trials': self.stim.v_trial.astype(int)
        }).unique(maintain_order=True)
        return df

    @property
    def n_trials(self) -> int:
        return self.unique_stimuli_set.shape[0]

    @property
    def i_stim(self) -> np.ndarray:
        """(N, ) value: stim type index"""
        return self.unique_stimuli_set.get_column('i_stims').to_numpy()

    @property
    def i_trial(self) -> np.ndarray:
        """(N, ). value: trial"""
        return self.unique_stimuli_set.get_column('i_trials').to_numpy()

    def get_time_interval(self) -> np.ndarray:
        ustims = self.stim.v_stim * (1 - self.stim.v_blank)
        utrials = self.stim.v_trial * (1 - self.stim.v_blank)

        ret = np.zeros([self.n_trials, 2])
        for i, (st, tr) in enumerate(self.unique_stimuli_set.iter_rows()):
            idx = np.where((ustims == st) & (utrials == tr))[0]
            ret[i, :] = self.stim.v_duino_time[[idx[0], idx[-1]]]

        return ret


# ======== #
# Protocol #
# ======== #

class PyVProtocol(AbstractStimProtocol):
    """
    class for handle the protocol file for pyvstim version (vb lab legacy)

    `Dimension parameters`:

        N = number of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = number of Stim Type

        C = number of Cycle
    """

    @classmethod
    def load(cls, file: Path | str, *,
             cast_numerical_opt=True) -> 'PyVProtocol':

        file = Path(file)
        options = {}
        version = 'pyvstim'

        state = 0
        with Path(file).open() as f:
            for line in f:
                line = line.strip()

                if len(line) == 0 or line.startswith('#'):
                    continue

                if state == 0 and line.startswith('n\t'):
                    header = re.split(r'\t+| +', line)
                    value = [[] for _ in range(len(header))]
                    state = 1

                elif state == 0:
                    idx = line.index('=')
                    if cast_numerical_opt:
                        opt_value = try_casting_number(line[idx + 1:].strip())
                    else:
                        opt_value = line[idx + 1:].strip()

                    options[line[:idx].strip()] = opt_value

                elif state == 1:
                    parts = re.split(r'\t+| +', line, maxsplit=len(header))
                    rows = unfold_stimuli_condition(parts)

                    for r in rows:
                        r.remove('')  # pyvstim interesting problem
                        for i, it in enumerate(r):  # for each col
                            if it != '':
                                value[i].append(it)
                else:
                    raise RuntimeError('illegal state')

            assert len(header) == len(value)
            visual_stimuli = {
                field: value[i]
                for i, field in enumerate(header)
            }

            if 'Shuffle' not in options.keys():
                options['Shuffle'] = False

        return PyVProtocol(file.name, options, pl.DataFrame(visual_stimuli), version)

    @property
    def is_shuffle(self) -> bool:
        """TODO"""
        return False

    @property
    def background(self) -> float:
        """TODO"""
        return self.options.get('background', 0.5)

    @property
    def start_blank_duration(self) -> int:
        raise NotImplementedError('')

    @property
    def blank_duration(self) -> int:
        return self.options['BlankDuration']

    @property
    def trial_blank_duration(self) -> int:
        raise NotImplementedError('')

    @property
    def end_blank_duration(self) -> int:
        raise NotImplementedError('')

    @property
    def trial_duration(self) -> int:
        raise NotImplementedError('')

    @property
    def visual_duration(self) -> int:
        raise NotImplementedError('')

    @property
    def total_duration(self) -> int:
        raise NotImplementedError('')

    def get_loops_expr(self) -> ProtExpression:
        """parse and get the expression and loop number"""
        exprs = []
        n_cycles = []
        n_blocks = self.visual_stimuli_dataframe.shape[0]

        for row in self.visual_stimuli_dataframe.iter_rows():  # each row item_value
            for it in row:
                if isinstance(it, str):
                    if 'loop' in it:
                        match = re.search(r"loop\((.*),(\d+)\)", it)

                        if match:
                            exprs.append(match.group(1))
                            n_cycles.append(match.group(2))
                    else:
                        fprint('loop info not found, check prot file!', vtype='warning')
                        exprs.append('')
                        n_cycles.append(1)

        return ProtExpression(exprs, list(map(int, n_cycles)), n_blocks)


@dataclasses.dataclass
class ProtExpression:
    """
    `Dimension parameters`:

        B = number of block

        C = number of Cycle
    """

    expr: list[str]
    """expression"""
    n_cycles: list[int]
    """number of cycle. length number = B?, value equal to C"""
    n_blocks: int | None
    """number of prot value row (block) B"""

    def __post_init__(self):
        if (len(self.n_cycles) == 2 * self.n_blocks) and self._check_ncycles_foreach_block():
            self.n_cycles = self.n_cycles[::2]
        else:
            raise RuntimeError('')

    def _check_ncycles_foreach_block(self):
        """check if the ncycles are the same and duplicate for each block"""
        n = len(self.n_cycles)
        if n % 2 != 0:
            return False

        for i in range(0, n, 2):
            if self.n_cycles[i] != self.n_cycles[i + 1]:
                return False

        return True



from matplotlib.axes import Axes

__all__ = ['ax_log_setting']


def ax_log_setting(ax: Axes, **kwargs):
    """log scale and tick setting"""
    import matplotlib.ticker as mticker

    ax.set_xscale('log', base=2)
    ax.set_yscale('log', base=2)
    ax.set_yticks([0.04, 0.08, 0.16])  # hardcode that used for interpolation plotting
    ax.set_xticks([1, 2, 4])  # hardcode that used for interpolation plotting
    ax.xaxis.set_major_formatter(mticker.ScalarFormatter())
    ax.xaxis.get_major_formatter().set_scientific(False)
    ax.xaxis.get_major_formatter().set_useOffset(False)
    ax.yaxis.set_major_formatter(mticker.ScalarFormatter())
    ax.yaxis.get_major_formatter().set_scientific(False)
    ax.yaxis.get_major_formatter().set_useOffset(False)
    ax.set_aspect(1.0 / ax.get_data_ratio(), adjustable='box')  # no matter xy scaling, set square

    # against the `plot_figure` ctx manager
    ax.spines['top'].set_visible(True)
    ax.spines['right'].set_visible(True)
    for axis in ['bottom', 'top', 'left', 'right']:
        ax.spines[axis].set_linewidth(1)

    ax.set(**kwargs)



import abc
from pathlib import Path

import cv2
import numpy as np

from neuralib.argp import AbstractParser, argument

__all__ = ['AbstractSegmentationOption']


class AbstractSegmentationOption(AbstractParser, metaclass=abc.ABCMeta):
    DESCRIPTION = 'ABC for cellular segmentation'

    file: Path = argument('-F', '--file', required=True, help='image file path')
    model: str = argument('-M', '--model', help='which pretrained model')
    napari_view: bool = argument('--napari', help='view in napari')

    no_normalize: bool = argument('--no-norm', help='NOT DO Percentile-based image normalization for eval')

    def load_gray_scale(self, normalize: bool = True) -> np.ndarray:
        if not self.file.is_file():
            raise ValueError('not a file')

        img = cv2.imread(str(self.file))
        if img.ndim == 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

        if normalize:
            from csbdeep.utils import normalize
            img = normalize(img, clip=True)

        return img

    def get_raw_image(self) -> np.ndarray:
        pass

    @abc.abstractmethod
    def eval(self):
        """eval the model"""
        pass

    @abc.abstractmethod
    def launch_napari(self, **kwargs):
        """napari viewer"""
        pass



from __future__ import annotations

import warnings
from contextlib import contextmanager
from pathlib import Path
from typing import Literal

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes

__all__ = ['plot_figure',
           'ax_set_default_style',
           'ax_merge']

MPL_BACKEND_TYPE = Literal[
    'GTK3Agg', 'GTK3Cairo', 'GTK4Agg', 'GTK4Cairo', 'MacOSX', 'nbAgg',
    'QtAgg', 'QtCairo', 'Qt5Agg', 'Qt5Cairo',
    'TkAgg', 'TkCairo',
    'WebAgg',
    'WX', 'WXAgg', 'WXCairo'
]


@contextmanager
def plot_figure(output: Path | None,
                *args,
                set_square: bool = False,
                set_equal_scale: bool = False,
                win_backend: MPL_BACKEND_TYPE = 'QtCairo',
                dpi: int | None = None,
                use_default_style: bool = True,
                **kwargs) -> Axes:
    """

    **Example**

    >>> fig_output = Path('output.png')
    >>> with plot_figure(fig_output) as ax:
    ...     ax.plot([0, 10], [0, 10])

    generate output.png

    :param output: output figure path
    :param args: plt.subplots(*args)
    :param set_square: sqaure plot
    :param set_equal_scale: equal for xy scaling. i.e., annotomical plot
    :param win_backend: handle backend problem in win10 pdf output.
        if high resolution image (WXAgg), otherwise keep normal pdf output (WXCario)
    :param dpi
    :param use_default_style:
    :param kwargs: plt.subplots(**kwargs)
    :return:
    """
    _os_handler(win_backend=win_backend)
    _mplrc_set()

    fig, ax = plt.subplots(*args, **kwargs)

    #
    if use_default_style:
        if isinstance(ax, np.ndarray):
            for _ax in ax.ravel():
                ax_set_default_style(_ax, set_square=set_square, set_equal_scale=set_equal_scale)
        else:
            ax_set_default_style(ax, set_square=set_square, set_equal_scale=set_equal_scale)
    #
    try:
        yield ax
    except:
        raise
    else:
        if output is None:
            plt.tight_layout()
            plt.show()
        else:
            while True:
                try:
                    # UserWarning: This figure includes Axes that are not compatible with tight_layout,
                    # so results might be incorrect
                    with warnings.catch_warnings():
                        warnings.simplefilter('ignore')
                        plt.tight_layout()
                except:
                    pass

                try:
                    plt.savefig(output, dpi=dpi if dpi is not None else None)
                    break
                except OSError as e:
                    print(e)
                    input('press to continue')

    finally:
        plt.clf()
        plt.close('all')


def _os_handler(win_backend: MPL_BACKEND_TYPE = 'WXCairo'):
    import platform
    import matplotlib as mpl

    if platform.system() == 'Windows':
        mpl.use(win_backend)


def _mplrc_set():  # TODO read from rc file with plt.rc_context()
    plt.rcParams['font.sans-serif'] = "Arial"


def ax_set_default_style(ax: Axes,  # TODO read from rc file with plt.rc_context()
                         set_square=False,
                         set_equal_scale=False):
    if 'polar' in ax.spines.keys():
        pass
    else:
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.xaxis.set_tick_params(width=1)
        ax.yaxis.set_tick_params(width=1)
        for axis in ['bottom', 'left']:
            ax.spines[axis].set_linewidth(1)

    if set_square:
        ax.set_aspect(1.0 / ax.get_data_ratio(), adjustable='box')

    if set_equal_scale:
        ax.set_aspect('equal')


# ========= #
# AxesMerge #
# ========= #


class AxesMergeHelper:
    def __init__(self, ax: np.ndarray):
        self.__fig = ax.ravel()[0].figure
        self.__gs = ax.ravel()[0].get_gridspec()
        self.__ax = ax

    def __getitem__(self, item) -> Axes:
        for ax in self.__ax[item].ravel():
            ax.set_visible(False)
            # ax.remove()

        ret = self.__fig.add_subplot(self.__gs[item])
        ax_set_default_style(ret)
        return ret


def ax_merge(ax: np.ndarray) -> AxesMergeHelper:  # TODO give doc with examples
    return AxesMergeHelper(ax)



from __future__ import annotations

from typing import Generic, TypeVar, final, Literal

import attrs
import cv2
import numpy as np
from typing_extensions import Self

from neuralib.util.util_type import PathLike

__all__ = [
    'ImageProcFactory',
    'IMAGE_CHANNEL_TYPE',
    #
    'apply_transformation',
    'detect_feature'
]

T = TypeVar('T')
IMAGE_CHANNEL_TYPE = Literal['red', 'green', 'blue', 'r', 'g', 'b']


@final
@attrs.define(repr=False)
class ImageProcFactory(Generic[T]):
    """Factory for basic imaging processing"""

    image: np.ndarray
    """(H, W, 4 | 3) or (H, W)"""

    color_type: T | None = attrs.field(init=False, default=None)
    """image color type"""

    def __repr__(self):
        shape = self.image.shape
        ctype = self.color_type
        return f'Shape: {shape}, color: {ctype}'

    def __attrs_post_init__(self):
        if self.image.ndim == 2:
            self.color_type = 'GrayScale'
        elif self.image.ndim == 3:
            c = self.image.shape[2]
            if c == 3:
                self.color_type = 'RGB'
            elif c == 4:
                self.color_type = 'RGBA'

        assert self.color_type is not None, 'color type could not be inferred'

    @classmethod
    def load(cls, file: PathLike,
             to_rgba: bool | None = None,
             to_rgb: bool | None = True) -> Self:

        img = cv2.imread(str(file))

        if to_rgba:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGBA)
            img = np.array(img, dtype=np.uint8)

        if to_rgb:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        return ImageProcFactory(img)

    @property
    def height(self) -> int:
        return self.image.shape[0]

    @property
    def width(self) -> int:
        return self.image.shape[1]

    def with_channel(self, channel: IMAGE_CHANNEL_TYPE) -> Self:
        """select rgb channel"""

        if self.color_type not in ('RGBA', 'RGB'):
            raise TypeError(f'{self.color_type} not able to split channel')

        if channel in ('r', 'red'):
            img = cv2.split(self.image)[0]
        elif channel in ('g', 'green'):
            img = cv2.split(self.image)[1]
        elif channel in ('b', 'blue'):
            img = cv2.split(self.image)[2]
        else:
            raise KeyError(f'{channel}')

        return attrs.evolve(self, image=img)

    def view_2d(self, flip: bool = True) -> Self:
        """view image as 2d array (H, W)"""
        if self.color_type == 'RGBA':
            w, h, _ = self.image.shape
            img = self.image.view(dtype=np.uint32).reshape((w, h))
        elif self.color_type == 'RGB':
            r = self.image[:, :, 0]
            g = self.image[:, :, 1]
            b = self.image[:, :, 2]
            grayscale_image = 0.2989 * r + 0.5870 * g + 0.1140 * b
            img = grayscale_image.astype(np.uint8)
        else:
            raise RuntimeError(f'{self.image.shape} check')

        if flip:
            img = np.flipud(img)

        return attrs.evolve(self, image=img)

    # ============= #
    # Basic Process #
    # ============= #
    def covert_grey_scale(self) -> Self:
        if self.color_type == 'GrayScale':
            return attrs.evolve(self, image=np.uint8(self.image))
        else:
            img = cv2.cvtColor(self.image, cv2.COLOR_BGR2GRAY)
            return attrs.evolve(self, image=img)

    def gaussian_blur(self, ksize: int, sigma: int) -> Self:
        img = cv2.GaussianBlur(self.image, (ksize, ksize), sigmaX=sigma, sigmaY=sigma)
        return attrs.evolve(self, image=img)

    def edge_detection(self,
                       lower_threshold: int = 30,
                       upper_threshold: int = 150) -> Self:
        grey_img = self.covert_grey_scale().image
        img = cv2.Canny(grey_img, lower_threshold, upper_threshold)
        return attrs.evolve(self, image=img)

    def binarize(self, threshold: int = 150) -> Self:
        _, img = cv2.threshold(self.image, threshold, 255, cv2.THRESH_BINARY)
        return attrs.evolve(self, image=img)

    def de_noise(self) -> Self:
        gray_img = self.covert_grey_scale().image
        return attrs.evolve(self, image=de_noise(gray_img))

    def local_maxima_image(self, channel: IMAGE_CHANNEL_TYPE, **kwargs) -> Self:
        """
        find the local maxima of the selection points.
        i.e., used in roi selection of the neuron before counting

        :param channel: color of image
        :return:
            (H, W) 2d array
        """
        from skimage.morphology import local_maxima

        image = self.with_channel(channel).image
        if np.sum(image) == 0:
            return attrs.evolve(self, image=np.zeros_like(image, dtype=np.uint8))
        else:
            return attrs.evolve(self, image=local_maxima(image, **kwargs))


def de_noise(gray_img: np.ndarray) -> np.ndarray:
    # https://stackoverflow.com/questions/62042172/how-to-remove-noise-in-image-opencv-python
    blur = cv2.GaussianBlur(gray_img, (3, 3), sigmaX=30, sigmaY=30)
    divide = cv2.divide(gray_img, blur, scale=255)
    # otsu threshold
    thresh = cv2.threshold(divide, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)[1]

    # apply morphology
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (3, 3))
    morph = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, kernel)

    cv2.imwrite("test.jpg", thresh)

    return thresh


def recover_overexposure(img: np.ndarray,
                         alpha: float = 0.5,
                         beta: float = 0.5) -> np.ndarray:
    """recover saturated fluorescence image"""
    proc = cv2.addWeighted(img, alpha, np.zeros_like(img), beta, 0.0)
    return proc


def apply_transformation(img: np.ndarray,
                         trans_mtx: np.ndarray,
                         **kwargs) -> np.ndarray:
    """2D image transform"""

    height, width, _ = img.shape

    if trans_mtx.shape != (3, 3):
        raise ValueError(f'invalid transformation shape: {trans_mtx.shape}')

    return cv2.warpPerspective(img, trans_mtx, (width, height), **kwargs)


def detect_feature(img: np.ndarray,
                   filter_type: str = 'ORB') -> tuple[np.ndarray, np.ndarray]:
    if filter_type == 'ORB':
        f = cv2.ORB_create()
    elif filter_type == 'SIFT':
        f = cv2.SIFT_create()
    else:
        raise NotImplementedError('')
    keypoints, descriptors = f.detectAndCompute(img, None)

    return keypoints, descriptors



from __future__ import annotations

import abc
from pathlib import Path
from typing import NamedTuple, Final, Any, Literal

import attrs
import numpy as np
import polars as pl
from scipy.io import loadmat
from scipy.io.matlab import MatlabOpaque

from neuralib.atlas.allen import AllenReferenceWrapper
from neuralib.atlas.util import PLANE_TYPE
from neuralib.atlas.view import SlicePlane
from neuralib.util.util_type import PathLike
from neuralib.util.util_verbose import fprint
from neuralib.util.utils import uglob, joinn

__all__ = [
    'AbstractCCFDir',
    'CCFBaseDir',
    'CCFOverlapDir',
    #
    'CCFTransMatrix',
    'load_transform_matrix'
]

# ======================== #
# CCF DataFolder Structure #
# ======================== #

CCF_GLOB_TYPE = Literal[
    'zproj',
    'roi',
    'roi_cpose',
    'resize',
    'resize_overlap',
    'processed',
    'transformation_matrix',
    'transformation_img',
    'transformation_img_overlap'
]

CHANNEL_SUFFIX = Literal['r', 'g', 'b', 'merge', 'overlap']


class AbstractCCFDir(metaclass=abc.ABCMeta):
    """
    ::

        ANIMAL_001/ (root)
            ├── raw/ (optional)
            ├── zproj/
            │    └── ANIMAL_001_g*_s*_{channel}.tif
            ├── roi/
            │    └── ANIMAL_001_g*_s*_{channel}.roi
            ├── roi_cpose/
            │    └── ANIMAL_001_g*_s*_{channel}.roi
            ├── resize/ (src for the allenccf)
            │    ├── ANIMAL_001_g*_s*_resize.tif
            │    └── processed/
            │           ├── ANIMAL_001_g*_s*_resize_processed.tif
            │           └── transformations/
            │                 ├── ANIMAL_001_g*_s*_resize_processed_transformed.tif
            │                 ├── ANIMAL_001_g*_s*_resize_processed_transform_data.mat
            │                 └── labelled_regions/
            │                       ├── {*channel}_roitable.csv
            │                       └── parsed_data /
            │                             └── parsed_csv_merge.csv
            │
            └── output_files/ (for generate output fig)

    """
    def __new__(cls, root: PathLike,
                auto_mkdir: bool = True,
                with_overlap_sources: bool = True):

        if with_overlap_sources:
            return object.__new__(CCFOverlapDir)
        else:
            return object.__new__(CCFBaseDir)

    def __init__(self, root: PathLike,
                 auto_mkdir: bool = True,
                 with_overlap_sources: bool = True):
        """
        :param root:
        :param auto_mkdir:
        :param with_overlap_sources:
        """
        self.root: Final[Path] = root

        if auto_mkdir:
            self._init_folder_structure()

        self.with_overlap_sources: Final[bool] = with_overlap_sources

    @abc.abstractmethod
    def _init_folder_structure(self) -> None:
        pass

    @abc.abstractmethod
    def glob(self,
             glass_id: int,
             slice_id: int,
             glob_type: CCF_GLOB_TYPE,
             hemisphere: Literal['i', 'c'] | None = None,
             channel: CHANNEL_SUFFIX | None = None) -> Path:
        """

        :param glass_id: Glass slide number
        :param slice_id: Slice sequencing under a glass slide
            (i.e., zigzag from upper left -> bottom left ->-> bottom right)
        :param glob_type:
        :param hemisphere:
        :param channel:
        :return:
        """
        pass

    @abc.abstractmethod
    def get_transformation_matrix(self,
                                  glass_id: int,
                                  slice_id: int,
                                  plane_type: PLANE_TYPE) -> CCFTransMatrix:
        pass


class CCFBaseDir(AbstractCCFDir):
    """Base folder structure for 2dccf pipeline"""

    def _init_folder_structure(self):
        iter_dir = (
            self.raw_folder,
            self.resize_folder,
            self.roi_folder,
            self.cpose_roi_folder,
            self.zproj_folder,
            self.processed_folder,
            self.transformed_folder,
            self.labelled_roi_folder,
            self.parsed_data_folder,
            self.output_folder,
        )

        for d in iter_dir:
            if not d.exists():
                d.mkdir(parents=True, exist_ok=True)
                fprint(f'auto make folder <{d.name}> for {self.animal}', vtype='io')

    @property
    def animal(self) -> str:
        return self.root.name

    @property
    def raw_folder(self) -> Path:
        return self.root / 'raw'

    @property
    def resize_folder(self) -> Path:
        return self.root / 'resize'

    @property
    def roi_folder(self) -> Path:
        return self.root / 'roi'

    @property
    def cpose_roi_folder(self) -> Path:
        return self.root / 'roi_cpose'

    @property
    def zproj_folder(self) -> Path:
        return self.root / 'zproj'

    # ========================================== #
    # CCF folder (MATLAB pipeline auto-generate) #
    # ========================================== #

    @property
    def processed_folder(self) -> Path:
        return self.resize_folder / 'processed'

    @property
    def transformed_folder(self) -> Path:
        return self.processed_folder / 'transformations'

    @property
    def labelled_roi_folder(self) -> Path:
        return self.transformed_folder / 'labelled_regions'

    @property
    def parsed_data_folder(self) -> Path:
        return self.labelled_roi_folder / 'parsed_data'

    @property
    def parse_csv(self) -> Path:
        return uglob(self.parsed_data_folder, 'parsed_csv_merge.csv')

    # ======= #
    # Outputs #
    # ======= #

    @property
    def output_folder(self) -> Path:
        return self.root / 'output_files'

    def figure_output(self, *suffix, sep='_') -> Path:
        ret = self.output_folder / joinn(sep, *suffix)
        return ret.with_suffix('.pdf')

    def csv_output(self, *suffix, sep='_') -> Path:
        ret = self.output_folder / joinn(sep, *suffix)
        return ret.with_suffix('.csv')

    # ========= #
    # File Glob #
    # ========= #

    def get_slice_id(self, glass_id: int,
                     slice_id: int,
                     hemisphere: Literal['i', 'c'] | None = None,
                     channel: CHANNEL_SUFFIX | None = None) -> str:

        ret = f'{self.animal}_{glass_id}_{slice_id}'
        if hemisphere is not None:
            ret += f'_{hemisphere}'
        if channel is not None:
            ret += f'_{channel}'
        return ret

    slice_name: str = None  # assign when glob

    def glob(self,
             glass_id: int,
             slice_id: int,
             glob_type: CCF_GLOB_TYPE, *,
             hemisphere: Literal['i', 'c'] | None = None,
             channel: CHANNEL_SUFFIX | None = None) -> Path | None:

        if glob_type in ('roi', 'roi_cpose', 'zproj'):
            self.slice_name = name = self.get_slice_id(glass_id, slice_id, hemisphere=hemisphere, channel=channel)
        else:
            self.slice_name = name = self.get_slice_id(glass_id, slice_id, hemisphere=hemisphere)

        #
        if glob_type == 'roi':
            return uglob(self.roi_folder, f'{name}*.roi')
        elif glob_type == 'roi_cpose':
            return uglob(self.cpose_roi_folder, f'{name}*cpose.roi')
        elif glob_type == 'zproj':
            return uglob(self.zproj_folder, f'{name}.*')
        elif glob_type == 'resize':
            return uglob(self.zproj_folder, f'{name}_resize.*')
        elif glob_type == 'transformation_matrix':
            return uglob(self.transformed_folder, f'{name}*.mat')
        elif glob_type == 'transformation_img':
            return uglob(self.transformed_folder, f'{name}*.tif')
        else:
            return

    def get_transformation_matrix(self,
                                  glass_id: int,
                                  slice_id: int,
                                  plane_type: PLANE_TYPE) -> CCFTransMatrix:
        return load_transform_matrix(
            self.glob(glass_id, slice_id, 'transformation_matrix'),
            plane_type
        )


class CCFOverlapDir(CCFBaseDir):
    """ 2dccf Folder structure for multiple sources overlap labeling
    For example: Dual tracing with different fluorescence protein, and tend to see the overlap channel counts"""

    def _init_folder_structure(self):
        super()._init_folder_structure()

        iter_overlap = (
            self.resize_overlap_folder,
            self.processed_folder,
            self.transformed_folder_overlap
        )
        for d in iter_overlap:
            if not d.exists():
                d.mkdir(parents=True, exist_ok=True)
                fprint(f'auto make folder <{d.name}> for {self.animal}', vtype='io')

    @property
    def resize_overlap_folder(self) -> Path:
        """for double lapping channel
        since maximal 2 channels for roi detection,
        then need extra folder use pseudo-color
        """
        return self.root / 'resize_overlap'

    @property
    def processed_folder_overlap(self) -> Path:
        return self.resize_overlap_folder / 'processed'

    @property
    def transformed_folder_overlap(self) -> Path:
        return self.processed_folder_overlap / 'transformations'

    def glob(self,
             glass_id: int,
             slice_id: int,
             glob_type: CCF_GLOB_TYPE, *,
             hemisphere: Literal['i', 'c'] | None = None,
             channel: CHANNEL_SUFFIX | None = None) -> Path | None:

        ret = super().glob(glass_id, slice_id, glob_type, hemisphere=hemisphere, channel=channel)

        if ret is None:
            if glob_type == 'resize_overlap':
                return uglob(self.resize_overlap_folder, f'{self.slice_name}_resize_overlap.*')
            elif glob_type == 'transformation_img_overlap':
                return uglob(self.transformed_folder_overlap, f'{self.slice_name}*.tif')

        return


class CCFSagittalDir(AbstractCCFDir):
    # TODO
    pass


# ===================== #
# Transformation matrix #
# ===================== #

class MatMatrix(NamedTuple):
    allen_location: np.ndarray
    """abnormal parsing in loadmat. [996 array([  0, -12], dtype=int16)]"""
    transform: MatlabOpaque
    """mat meta info. currently not used"""
    transform_points: np.ndarray
    """[(P, 2), (P, 2)] transformation point for 2dccf GUI site view"""

    @property
    def slice_index(self) -> int:
        return int(self.allen_location[0])

    # noinspection PyTypeChecker
    @property
    def angle_xy(self) -> tuple[float, float]:
        """angle shifting for xy axis (width, height) for a specific plane type"""
        angle = self.allen_location[1]
        return angle[0], angle[1]


def load_transform_matrix(filepath: PathLike,
                          plane_type: PLANE_TYPE,
                          resolution: int = 10,
                          default_name: str = 'test') -> CCFTransMatrix:
    """
    matrix for image transformation

    :param filepath:
    :param plane_type
    :param resolution
    :param default_name
    :return:
    """
    mat = loadmat(filepath, squeeze_me=True, struct_as_record=False)['save_transform']
    f = Path(filepath).name

    try:
        name = f[:f.index('_resize')]
    except ValueError:
        name = default_name

    return CCFTransMatrix(name,
                          MatMatrix(
                              mat.allen_location,
                              mat.transform,
                              mat.transform_points
                          ),
                          plane_type,
                          resolution=resolution)


@attrs.frozen
class CCFTransMatrix:
    slice_id: str
    matrix: MatMatrix
    plane_type: PLANE_TYPE
    resolution: int = attrs.field(kw_only=True, default=10)

    @property
    def slice_index(self) -> int:
        return self.matrix.slice_index

    @property
    def angle_xy(self) -> tuple[float, float]:
        return self.matrix.angle_xy

    def get_slice_plane(self) -> SlicePlane:
        ret = (AllenReferenceWrapper
               .load_slice_view('template', self.plane_type, self.resolution)
               .plane_at(self.slice_index))

        if np.any([a != 0 for a in self.matrix.angle_xy]):
            ret = ret.with_angle_offset(-self.matrix.angle_xy[0], self.matrix.angle_xy[1])

        return ret

    def matrix_info(self, to_polars: bool = True) -> dict[str, Any] | pl.DataFrame:
        ret = {
            'slice_id': self.slice_id,
            'plane_type': self.plane_type,
            'atlas_resolution': self.resolution,
            'slice_index': self.slice_index,
            'reference_value': self.get_slice_plane().reference_value,
            'dw': self.angle_xy[0],
            'dh': self.angle_xy[1]
        }

        if to_polars:
            ret = pl.DataFrame(ret)

        return ret



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import Literal, Union

import pandas as pd
import polars as pl
from colorama import Fore, Style

from neuralib.util.util_type import DataFrame

__all__ = ['fprint',
           'printdf',
           'print_load',
           'print_save']

_PREV_LOAD_FILE = None
_PREV_SAVE_FILE = None


def fprint(*msgs,
           vtype: Literal['info', 'io', 'warning', 'error', 'pass'] = 'info',
           timestamp: bool = True,
           **kwarg) -> None:
    """
    Formatting print with different colors based on verbose type

    :param msgs:
    :param vtype: verbose type
    :param timestamp:
    :return:
    """

    if vtype == 'error':
        prefix = '[ERROR]'
        color = 'red'
    elif vtype == 'warning':
        prefix = '[WARNING] '
        color = 'yellow'
    elif vtype == 'io':
        prefix = '[IO] '
        color = 'magenta'
    elif vtype == 'info':
        prefix = '[INFO]'
        color = 'cyan'
    elif vtype == 'pass':
        prefix = '[PASS]'
        color = 'green'
    else:
        raise ValueError(f'{vtype}')

    try:
        fg_color = getattr(Fore, color.upper())
    except AttributeError:
        fg_color = Fore.WHITE

    msg = fg_color + prefix
    if timestamp:
        msg += f"[{datetime.today().strftime('%y-%m-%d %H:%M:%S')}] - "

    try:
        out = f"{''.join(msgs)}\n"
    except TypeError:
        out = f'{msgs}'

    msg += out
    msg += Style.RESET_ALL
    print(msg, **kwarg)


def printdf(df: DataFrame,
            nrows: int | None = None,
            ncols: int | None = None,
            tbl_width_chars: int = 500,
            do_print: bool = True) -> str:
    """
    print dataframe with given row numbers (polars)
    if isinstance pandas dataframe, print all.

    :param df: polars or pandas dataframe
    :param nrows: number of rows (applicable in polars case)
    :param ncols: number of columns
    :param tbl_width_chars: table width for showing
    :param do_print: do print otherwise, only return the str
    :return:
    """

    if isinstance(df, pl.DataFrame):
        with pl.Config(tbl_width_chars=tbl_width_chars) as cfg:
            rows = df.shape[0] if nrows is None else nrows
            cols = df.shape[1] if ncols is None else ncols
            cfg.set_tbl_rows(rows)
            cfg.set_tbl_cols(cols)

            if do_print:
                print(df)

            return df.__repr__()

    elif isinstance(df, pd.DataFrame):
        ret = df.to_markdown()
        print(ret)
        return ret

    else:
        raise TypeError('')


def print_load(file: Union[str, Path], verb='LOAD') -> Path:
    """print message for loading file.

    If *file* is not existed, '!' will add after *verb*.
    File load message doesn't print twice and more if it shows continuously, so it is okay
    that wrap a loading method and call :func:`print_load` on target file.

    use flag "load".

    :param file: file path
    :param verb: default 'LOAD'
    :param color: verb color
    :return: *file*
    """
    global _PREV_LOAD_FILE

    file = Path(file)

    if _PREV_LOAD_FILE != file:
        if len(verb) < 4:
            verb += ' ' * (4 - len(verb))

        not_exist = '!' if not file.exists() else ' '

        from tqdm import tqdm
        with tqdm.external_write_mode():
            fprint(f'{verb}{not_exist}->{file}', vtype='io')

    _PREV_LOAD_FILE = file
    return file


def print_save(file: Union[str, Path], verb='SAVE') -> Path:
    """print message for saving file.

    If *file* is not existed, '+' will add after *verb*.
    File save message doesn't print twice and more if it shows continuously, so it is okay
    that wrap a saving method and call :func:`print_load` on target file.

    use flag "save".

    :param file: file path
    :param verb: default 'SAVE'
    :param color: verb color
    :return: *file*
    """
    global _PREV_SAVE_FILE

    file = Path(file)

    if _PREV_SAVE_FILE != file:
        if len(verb) < 4:
            verb += ' ' * (4 - len(verb))

        not_exist = '+' if not file.exists() else ' '

        from tqdm import tqdm
        with tqdm.external_write_mode():
            fprint(f'{verb}{not_exist}->{file}', vtype='io')

    _PREV_SAVE_FILE = file
    return file



from __future__ import annotations

import attrs
import polars as pl
from typing_extensions import Self, Final, final

from neuralib.atlas.ccf.classifier import supply_overlap_dataframe
from neuralib.atlas.type import Source, Area
from neuralib.atlas.util import get_margin_merge_level

__all__ = ['RoiAreaQuery', 'SubregionResult']

from neuralib.util.util_verbose import fprint


@final
class RoiAreaQuery:
    """class for finding a subset of rois from a specific area"""

    def __init__(self,
                 parsed_df: pl.DataFrame,
                 area: Area,
                 source_order: tuple[Source, ...] | None = None,
                 force_set_show_col_level: int | None = None):
        """
        :param parsed_df: parsed dataframe from ccf pipeline
        :param area: area name
        :param source_order: order of the unique sources
        :param force_set_show_col_level: force set show col to which level.
            use case: if a low level area name is classified and show in high level (i.e., TH).
            Then directly specify the level number instead of auto inferred by hierarchical_query()

            **Please use carefully**

            Note that it results in subregions mixed in different level

            for example, if one level contain both HIP and HPF, it will quantify and calculate the percentage together,
            however, they are not in the same level for actual allen tree level. it's due to the raw data issue
            (Rois are not classified correctly initially in allenCCF pipeline)
        """

        self.parsed_df: Final[pl.DataFrame] = parsed_df
        self.area: Final[Area] = area

        if source_order is None:
            self.source_order = tuple(self.parsed_df['source'].unique().to_list())
        else:
            self.source_order = source_order

        self._query_col = get_margin_merge_level(self.parsed_df, self.area, 'lowest')

        # infer after query if not force set
        self._show_col: str | None = None if force_set_show_col_level is None else f'merge_ac_{force_set_show_col_level}'
        self.query_result: Final[pl.DataFrame] = self._hierarchical_query()

    def __repr__(self):
        q = f'Query -> {self.query_col}'
        s = f'Show -> {self.show_col}'
        res = self.query_result.select('name', self.query_col, self.show_col)
        return '\n'.join([q, s, str(res)])

    @classmethod
    def by(cls, df: pl.DataFrame,
           area: Area, *,
           source_order: tuple[Source, ...] | None = None,
           force_set_show_col_level: int | None = None) -> Self:
        """
        Query which area

        :param df: ccf parsed dataframe
        :param area: area name
        :param source_order: order of the unique sources
        :param force_set_show_col_level: force set show col to which level.
            use case: if a low level area name is classified and show in high level (i.e., TH).
            Then directly specify the level number instead of auto inferred by hierarchical_query()

            **Please use carefully**

            Note that it results in subregions mixed in different level

            for example, if one level contain both HIP and HPF, it will quantify and calculate the percentage together,
            however, they are not in the same level for actual allen tree level. it's due to the raw data issue
            (Rois are not classified correctly initially in allenCCF pipeline)
        :return:
        """

        return RoiAreaQuery(df, area, source_order=source_order, force_set_show_col_level=force_set_show_col_level)

    @property
    def query_col(self) -> str:
        """which column for the area name query (base column query source)"""
        return self._query_col

    @property
    def show_col(self) -> str:
        """highest merge level column for a given area (to show)"""
        return self._show_col

    def _hierarchical_query(self) -> pl.DataFrame:
        """Get the query dataframe
        Auto infer query col & show_col based on area name"""
        if self._show_col is None:
            highest_lv = int(get_margin_merge_level(self.parsed_df, self.area, 'highest').split('_')[-1])
            self._show_col = f'merge_ac_{highest_lv + 1}'

            if self._show_col not in self.parsed_df.columns:
                fprint(f'{self._show_col} exceed level, force show *acronym*', vtype='warning')
                self._show_col = 'acronym'

        return self.parsed_df.filter(pl.col(self.query_col) == self.area)

    def get_subregion_result(self, unit: str,
                             supply_overlap: bool = True) -> SubregionResult:
        """
        Get subregion results

        :param unit: {'n_rois', 'percent'}
        :param supply_overlap: add overlap roi counts into other channel(s)
        :return: :class:`SubregionResult`
        """
        dat = self.query_result.select(['source', self.show_col])

        if supply_overlap:
            dat = supply_overlap_dataframe(dat)

        df = (dat.group_by(['source', self.show_col])
              .agg(pl.col(self.show_col).count().alias('n_rois'))
              .with_columns((pl.col('n_rois') / pl.col('n_rois').sum().over('source') * 100).alias('percent'))
              .sort('percent', descending=True))

        idx = {val: idx for idx, val in enumerate(self.source_order)}
        ch_sort = pl.col('source').replace(idx)

        #
        numbers = df.group_by('source').agg(pl.col('n_rois').sum())
        numbers = numbers.sort(ch_sort)

        ch_count_all = self.parsed_df.group_by('source').len()
        roi_profile = (numbers.join(ch_count_all, on='source')
                       .with_columns((pl.col('n_rois') / pl.col('len')).alias('total_perc'))
                       .sort(ch_sort))
        #
        df = (df.pivot(columns=self._show_col, index='source', values=unit, aggregate_function='first')
              .fill_null(0))

        df = df.sort(ch_sort)

        return SubregionResult(self, unit, df, roi_profile)


@attrs.define
class SubregionResult:
    source_query: RoiAreaQuery
    """source query"""

    unit: str = attrs.field(validator=attrs.validators.in_(('n_rois', 'percent')))
    """Note that percent are normalized by queried dataframe"""

    data: pl.DataFrame
    """
    * Optional with animal col::
    
        ┌─────────┬─────────┬─────────┬─────────┬─────────┬─────────┬─────────┬─────────┬─────────┬────────┐
        │ source  ┆ VISam   ┆ VISp    ┆ VISpm   ┆ VISl    ┆ VISal   ┆ VISpor  ┆ VISli   ┆ VISpl   ┆ VISC   │
        │ ---     ┆ ---     ┆ ---     ┆ ---     ┆ ---     ┆ ---     ┆ ---     ┆ ---     ┆ ---     ┆ ---    │
        │ str     ┆ f64     ┆ f64     ┆ f64     ┆ f64     ┆ f64     ┆ f64     ┆ f64     ┆ f64     ┆ f64    │
        ╞═════════╪═════════╪═════════╪═════════╪═════════╪═════════╪═════════╪═════════╪═════════╪════════╡
        │ aRSC    ┆ 25.4669 ┆ 34.0318 ┆ 23.0979 ┆ 6.42369 ┆ 6.37813 ┆ 1.77676 ┆ 2.27790 ┆ 0.0     ┆ 0.5466 │
        │         ┆ 7       ┆ 91      ┆ 5       ┆         ┆ 2       ┆ 5       ┆ 4       ┆         ┆ 97     │
        │ pRSC    ┆ 15.1445 ┆ 36.1079 ┆ 29.5568 ┆ 10.7129 ┆ 3.04431 ┆ 2.38921 ┆ 1.92678 ┆ 1.07899 ┆ 0.0385 │
        │         ┆ 09      ┆         ┆ 4       ┆ 09      ┆ 6       ┆         ┆ 2       ┆ 8       ┆ 36     │
        │ overlap ┆ 48.5294 ┆ 8.82352 ┆ 31.25   ┆ 4.04411 ┆ 3.30882 ┆ 2.57352 ┆ 1.10294 ┆ 0.0     ┆ 0.3676 │
        │         ┆ 12      ┆ 9       ┆         ┆ 8       ┆ 4       ┆ 9       ┆ 1       ┆         ┆ 47     │
        └─────────┴─────────┴─────────┴─────────┴─────────┴─────────┴─────────┴─────────┴─────────┴────────┘
    """

    roi_profile: pl.DataFrame
    """
    roi profile::
    
        ┌─────────┬────────┬───────┬────────────┐
        │ source  ┆ n_rois ┆ count ┆ total_perc │
        │ ---     ┆ ---    ┆ ---   ┆ ---        │
        │ str     ┆ u32    ┆ u32   ┆ f64        │
        ╞═════════╪════════╪═══════╪════════════╡
        │ aRSC    ┆ 899    ┆ 11403 ┆ 0.078839   │
        │ pRSC    ┆ 6744   ┆ 26280 ┆ 0.256621   │
        │ overlap ┆ 225    ┆ 3470  ┆ 0.064841   │
        └─────────┴────────┴───────┴────────────┘
    """

    def del_overlap(self) -> Self:
        """keep same `source_query` but delete the `overlap` data and roi_profile"""
        expr = pl.col('source') != 'overlap'
        dat = self.data.filter(expr)
        profile = self.roi_profile.filter(expr)
        return attrs.evolve(self, data=dat, roi_profile=profile)

    def with_animal(self, animal: str) -> Self:
        """add animal col for batch/statistic purpose"""
        data = self.data.with_columns(pl.lit(animal).alias('animal'))
        return attrs.evolve(self, data=data)

    @property
    def sources(self) -> list[Source]:
        """list of sources"""
        return self.data['source'].to_list()

    @property
    def channel_weight(self) -> pl.DataFrame:
        """fraction foreach source in all rois"""
        return self.roi_profile.drop('n_rois', 'count')

    @property
    def weight_list(self) -> list[float]:
        """list of perc"""
        return self.channel_weight.get_column('total_perc').to_list()

    @property
    def values(self) -> pl.DataFrame:
        """numerical data without source column"""
        return self.data.select(pl.exclude('source'))

    @property
    def areas(self) -> list[Area]:
        """list of areas"""
        return self.values.columns

    @property
    def n_areas(self) -> int:
        """number of areas"""
        return len(self.areas)

    @property
    def n_rois_info(self) -> list[str]:
        """info for plotting"""
        return [
            f'{self.source_query.area} -> {row[0]}: {row[1]} ({round(self.weight_list[i], 3)})'
            for i, row in enumerate(self.roi_profile.iter_rows())
        ]



from __future__ import annotations

from brainrender._colors import get_random_colors

__all__ = ['get_color']


def get_color(i, color_pattern: str | tuple[str, ...] | list[str]) -> str:
    """
    get color
    :param i: idx of the color pattern list
    :param color_pattern: color pattern list or single element str
    :return:
        color name
    """
    if isinstance(color_pattern, (list, tuple)):
        if i >= len(color_pattern):
            color_pattern.extend(get_random_colors(i + 2 - len(color_pattern)))
        return color_pattern[i]
    elif isinstance(color_pattern, str):
        return color_pattern



from functools import cached_property
from typing import Literal

import attrs
import napari
import numpy as np
from stardist.models import StarDist2D
from typing_extensions import Self

from neuralib.argp import argument, as_argument
from neuralib.segmentation.base import AbstractSegmentationOption

STARDIST_MODEL = Literal['2D_versatile_fluo', '2D_versatile_he', '2D_paper_dsb2018', '2D_demo']


@attrs.define
class StarDistResult:
    """
    Shape Info
    N: all cell numbers
    E: polygons edge numbers
    """
    labels: np.ndarray
    """image with label"""
    cord: np.ndarray
    """(N, 2, E) """
    prob: np.ndarray
    """(N,)"""

    point: np.ndarray = attrs.field(init=False)
    """coordinates to points by simple xy average"""

    def __attrs_post_init__(self):
        p = np.zeros((self.cord.shape[0], 2))
        for i, c in enumerate(self.cord):
            x_mean = np.mean(c[0, :])
            y_mean = np.mean(c[1, :])
            p[i, 0] = x_mean
            p[i, 1] = y_mean
        self.point = p

    def with_probability(self, threshold: float) -> Self:
        m = self.prob >= threshold
        return attrs.evolve(self, cord=self.cord[m], prob=self.prob[m])


class RunStarDist2DOptions(AbstractSegmentationOption):
    DESCRIPTION = 'Run the Stardist model for segmentation'

    model: STARDIST_MODEL = as_argument(AbstractSegmentationOption.model).with_options(
        default='2D_versatile_fluo',
        help='stardist pretrained model'
    )

    prob_thresh: float = argument('--prob', default=0.5,
                                  help='Consider only object candidates from pixels with predicted object probability '
                                       'above this threshold. '
                                       'Seealso: stardist.models.base._predict_instances_generator: prob_thresh')

    def run(self):
        if self.napari_view:
            self.launch_napari()

    @cached_property
    def raw_image(self) -> np.ndarray:
        return self.load_gray_scale(normalize=False)

    @property
    def normalized_images(self) -> np.ndarray:
        return self.load_gray_scale(normalize=True)

    def eval(self, prob: float = 0.5, **kwargs) -> StarDistResult:
        model = StarDist2D.from_pretrained(self.model)
        img = self.load_gray_scale()
        labels, detail = model.predict_instances(img, prob_thresh=self.prob_thresh, **kwargs)

        return StarDistResult(labels, detail['coord'], detail['prob']).with_probability(self.prob_thresh)

    def _remove_outlier(self):
        pass

    def launch_napari(self, with_widget=False):
        res = self.eval()
        viewer = napari.Viewer()
        viewer.add_image(self.raw_image, name='raw')
        viewer.add_image(self.normalized_images, name='normalized')
        viewer.add_image(res.labels, name='labels', colormap='cyan', opacity=0.5)
        viewer.add_points(res.point, face_color='red')

        if with_widget:
            viewer.window.add_plugin_dock_widget("stardist-napari", "StarDist")

        napari.run()


if __name__ == '__main__':
    RunStarDist2DOptions().main()



"""
Setup a color logger
=============

Example:
---------

.. code-block:: python

    from neuralib.util.color_logging import setup_clogger

    logger = setup_clogger()

    logger.debug("a debug message")
    logger.info("an info message")
    logger.warning("a warning message")
    logger.error("an error message")
    logger.critical("a critical message")

    # custom logging
    LOGGING_IO_LEVEL = 11
    LOGGING_IO_NAME = 'IO'

    logger.log(LOGGING_IO_LEVEL, 'io information message')

"""
from __future__ import annotations

import logging

__all__ = [
    'LOGGING_IO_LEVEL',
    #
    'setup_logger',
    'setup_clogger'
]

# custom
LOGGING_IO_LEVEL = 11
LOGGING_IO_NAME = 'IO'


def setup_logger() -> logging.Logger:
    """basic built-in logger"""
    _format = "%(asctime)s %(levelname)s %(message)s"
    logging.basicConfig(format=_format, level=logging.INFO)
    return logging.getLogger()


def setup_clogger(level: int | str = 10,
                  caller_name: str | None = None) -> logging.Logger:
    """
    Return a logger with a default ColoredFormatter.

    :param level: default level = 10 (DEBUG level)
    :param caller_name: show the name <Path name, Script name, ...> if needed
    :return:
    """
    from colorlog import ColoredFormatter

    logging.addLevelName(LOGGING_IO_LEVEL, LOGGING_IO_NAME)

    if caller_name is None:
        _format = "%(log_color)s %(levelname)-8s%(reset)s %(log_color)s[%(asctime)s] %(light_green)s %(message)s"
    else:
        _format = "%(log_color)s %(levelname)-8s%(reset)s %(log_color)s[%(asctime)s] [%(name)s] %(light_green)s %(message)s"

    formatter = ColoredFormatter(
        _format,
        datefmt=None,
        reset=True,
        log_colors={
            "DEBUG": "white",
            'IO': "purple",
            "INFO": "cyan",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red",
        },
    )

    logger = logging.getLogger(caller_name)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    #
    if isinstance(level, str) and level not in formatter.log_colors.keys():
        raise ValueError(f'invalid level string: {level}')
    logger.setLevel(level)

    return logger





"""
Confocal Scanner
======================

- This module provide the reading and parsing the Zeiss confocal dataset,
currently support .lsm & .czi format.

- Use for batch processing/analysis, or other api integration.
(i.e., batch cell segmentation, manually stitching...)

- For general image visualization, **Zen/Fiji/ImageJ** are recommended




Image Data Dim
---------------

`Dimension parameters (DimCode)`:

    V - view

    H - phase

    I - illumination

    S - scene

    R - rotation

    T - time

    C - channel

    Z - z plane (height)

    M - mosaic tile, mosaic images only

    Y - image height

    X - image width

    A - samples, BGR/RGB images only




LSM format
------------

.. code-block:: python

    from neuralib.scanner import LSMConfocalScanner

    filepath = ...
    lsm = LSMConfocalScanner.load(filepath)

    # get the meta
    print(lsm.meta)

    # get the images array
    print(lsm.lsmfile)

    # zproj imshow
    lsm.imshow(channel=2, zproj_type='max')



CZI format
------------

.. code-block:: python

    from neuralib.scanner import LSMConfocalScanner

    filepath = ...
    czi = CziConfocalScanner.load(filepath)

    # get meta
    print(czi.meta)

    # get dim code
    print(czi.get_dim_code())

    # get the image array
    print(czi.get_image(channel=0, scene=1))

    # zproj imshow
    lsm.imshow()

"""
from .core import *
from .czi import *
from .lsm import *



"""
Persistence Class
=================

:author:
    Ta-Shun Su

Define a persistence class
---------------------------

Import

>>> import numpy as np
>>> from neuralib import persistence

Define class

>>> @persistence.persistence_class
... class Example:
...     # key. use filename keyword to display key in filename.
...     use_animal: str = persistence.field(validator=True, filename=True)
...     use_session: str = persistence.field(validator=True)
...     use_date: str = persistence.field(validator=True, filename=True)
...     # data
...     channels: list[int]
...     data: np.ndarray

Load/Save

>>> example = Example(use_animal='A00', use_session='', use_date='1234')
>>> save(example, 'example.pkl')
>>> example_2 = load(Example, 'example.pkl') # example and example_2 should be content identical


Cooperate with PersistenceOptions
----------------------------------

>>> from neuralib.persistence.cli_persistence import PersistenceOptions
>>> class ExampleHandle(PersistenceOptions[Example]):
...     def empty_cache(self) -> Example:
...         return Example(use_animal='A00', use_session='', use_date='1234') # with attribute initialization
...     def compute_cache(self, result: Example) -> Example:
...         result.channels = [0, 1, 2]
...         result.riglog = np.array(result.channels)
...         return result

Dynamic generated methods for persistence class
------------------------------------------------

1. `__init__({foreach persistence.field})`

    >>> class Example:
    ...     ... # as same as above
    ...     def __init__(self, use_animal:str,  use_session: str, use_date: str): # auto generated
    ...         ...

2. `__str__` return filename

    >>> class Example:
    ...     ... # as same as above
    ...     def __str__(self): # auto generated
    ...         return filename(self)

3. `__repr__`

    >>> class Example:
    ...     ... # as same as above
    ...     def __repr__(self): # auto generated
    ...         return 'Example{' + f'use_animal={self.use_animal}, use_session={self.use_session}, use_date={self.use_date}' + "}"

4. `_replace` when persistence define empty `_replace` methods. It is NamedTuple._replace like function.

    >>> class Example:
    ...     ... # as same as above
    ...     def _replace(self, **kwargs): pass # empty method
    ...     def _replace(self, *,  # replaced by generated
    ...                  use_animal=missing,
    ...                  use_session=missing,
    ...                  use_date=missing,
    ...                  channels=missing,
    ...                  data=missing) -> Example:
    ...         ...

Auto increment field
---------------------

For some reason you want to save a persistence result that came from same data source but different
contents, which are usualy generated by random or suffle process. To save them all separately, you may
need a field value that keep track that it is n-th persistence result. :func:`autoinc_field` is proposed
to help this case.


>>> @persistence_class
... class Result:
...     a: str = field(validator=True, filename=T)
...     b: int = autoinc_field()
...     c: str
...     def __init__(self, a: str): # auto generated signature
...     def _replace(self, *, a: str, c:str): # auto generated signature

There are some rule when a persistence class has an autoinc_field.

1. only one autoinc_field in a persistence class allowed.
2. field type only int is allowed.
3. raise error when load a result without autoinc field resolved.
4. autoinc field is auto resolved when saving. Its value is max(found) + 1

Pickle format
--------------

Persistence class is transformed into a dict by `as_dict`, which as a root instance to be saved into
a pickle file.

**IMPORTANT**

If a persistance class has a custom `__init__` function which signature is differed from auto generated,
you need to define a classmethod `from_dict` for creating that persistance class.

>>> @persistence.persistence_class
... class Example:
...     a: int = persistence.field(validator=True, filename=True)
...     b: int = persistence.field(validator=True, filename=True)
...     c: int
...     def __init__(self): ... # custom __init__
...     @classmethod
...     def from_dict(cls, data:dict[str, Any]) -> 'Example':
...         # data = {'a', 'b', 'c'}


"""
from __future__ import annotations

import abc
import inspect
from pathlib import Path
from typing import Type, TypeVar, Union, Callable, Optional, Generic, Any, Iterable, get_type_hints

import numpy as np

from neuralib.util.func import create_fn
from neuralib.util.util_type import is_iterable
from neuralib.util.util_verbose import fprint

__all__ = [
    'field',
    'autoinc_field',
    'persistence_class',
    'ensure_persistence_class',
    'as_dict',
    'from_dict',
    'load',
    'save',
    'filename',
    'AutoIncFieldNotResolvedError',
    'auto_generated_content',
    'PersistenceHandler',
    'PickleHandler',
    'GzipHandler',
    #
    'ETLConcatable',
    'validate_concat_etl_persistence',
]

T = TypeVar('T')
P = Union[str, Path]

missing = inspect.Parameter.empty

VALIDATOR = Callable[[T, T], bool]


def field(validator: Union[bool, VALIDATOR] = False,
          filename_prefix: str = '',
          filename: Union[bool, Callable[..., str]] = False) -> T:
    """Cache class's exported field. Used as keys to find correspond persistence.

    :param validator: validate this field. use __eq__ by default. Can be a callable as a customized validator.
    :param filename_prefix: prefix word of *filename*
    :param filename: display this value on filename. Can be a callable that return the string.
    :return:
    """
    return PersistentField(validator, filename_prefix, filename, init=True)


def autoinc_field(filename_prefix: str = '') -> int:
    """make a field as auto increment, which can be used to save and distinguish between same source data but
    different random/shuffle persistence result.

    :param filename_prefix:
    :return:
    """
    # noinspection PyTypeChecker
    return PersistentField(validator=True, filename_prefix=filename_prefix, filename=True, init=False, autoinc=True)


class PersistentField(Generic[T]):
    """exported field of persistence class."""

    __slots__ = ('field_name', 'field_type', 'validator', 'filename_prefix', 'filename', 'init', 'autoinc')

    def __init__(self,
                 validator: Union[bool, VALIDATOR] = False,
                 filename_prefix: str = '',
                 filename: Union[bool, Callable[..., str]] = False,
                 init=True,
                 autoinc=False):
        """

        :param validator: validate this field. use __eq__ by default. Can be a callable as a customized validator.
        :param filename_prefix: prefix word of *filename*
        :param filename: display this value on filename. Can be a callable that return the string.
        :param init: put this field into class __init__
        """
        self.validator = validator
        self.filename_prefix = filename_prefix
        self.filename = filename
        self.init = init
        self.autoinc = autoinc

    def __set_name__(self, owner: Type, name: str):
        self.field_name = name
        if self.autoinc:
            if (field_type := get_type_hints(owner).get(name, Any)) != int:
                raise RuntimeError(f'type of autoinc field {name} should be int, but {field_type}')
        else:
            self.field_type = owner.__annotations__.get(name, Any)

    def validate(self, v: T, u: T) -> bool:
        """Validate this two value.

        :param v:
        :param u:
        :return:
        """
        if self.validator is False:
            return True
        elif self.validator is True:
            return v == u
        elif callable(self.validator):
            return self.validator(v, u)
        else:
            raise RuntimeError()


class AutoIncFieldNotResolvedError(RuntimeError):
    def __init__(self, instance, field: Union[str, PersistentField], message: str = None):
        if isinstance(field, PersistentField):
            field = field.field_name

        if message is None:
            message = f'{type(instance).__name__} autoinc field {field} is not resolved'

        super().__init__(message)

        self.instance = instance
        self.field = field


def persistence_class(cls: Type = None, /, *,
                      name: str = None,
                      filename_field_splitter='-'):
    """A class decorator.

    Decorated class ...

    :param cls: Persistence class.
    :param name: class name as filename.
    :param filename_field_splitter: the field splitter on filename.
    :return:
    """

    def decorator(cls: Type):
        cls._ast_persistence_cls_info_ = pc = PersistentClass(cls, name, filename_field_splitter)

        prev_autoinc_field = None
        for attr_name, attr_type in cls.__annotations__.items():
            if isinstance((attr_value := getattr(cls, attr_name, None)), PersistentField):
                delattr(cls, attr_name)

                f = attr_value
            else:
                f = PersistentField(init=False)
                f.field_name = attr_name
                f.field_type = attr_type

            if f.autoinc:
                if prev_autoinc_field is None:
                    prev_autoinc_field = f
                else:
                    raise RuntimeError('duplicated auto_inc_field')

            pc.fields.append(f)

        if cls.__init__ == object.__init__:
            cls.__init__ = _persistence_class_init(pc)
        if cls.__str__ == object.__str__:
            cls.__str__ = _persistence_class_str(pc)
        if cls.__repr__ == object.__repr__:
            cls.__repr__ = _persistence_class_repr(pc)
        if hasattr(cls, '_replace'):
            cls._replace = _persistence_class_replace(pc)
        return cls

    if cls is None:
        return decorator
    else:
        return decorator(cls)


class PersistentClass(Generic[T]):
    """persistence info class"""

    __slots__ = ('persistence_cls', 'cls_name', 'filename_field_splitter', 'fields')

    def __init__(self,
                 persistence_cls: Type,
                 persistence_name: str = None,
                 filename_field_splitter='-'):
        self.persistence_cls = persistence_cls
        self.cls_name = persistence_name if persistence_name is not None else persistence_cls.__name__
        self.filename_field_splitter = filename_field_splitter
        self.fields: list[PersistentField] = []

    def fields_name(self) -> list[str]:
        return [it.field_name for it in self.fields]

    def get_field(self, name: str) -> Optional[PersistentField]:
        for f in self.fields:
            if f.field_name == name:
                return f
        return None

    def autoinc_field(self) -> Optional[PersistentField]:
        for f in self.fields:
            if f.autoinc:
                return f
        return None

    def is_autoinc_field_resolved(self, result: Optional[T], **kwargs) -> bool:
        if (af := self.autoinc_field()) is None:
            return True
        return getattr(result, af.field_name, missing) is not missing or af.field_name in kwargs

    def validate(self, v: T, u: Union[T, dict[str, Any]]) -> bool:
        """validate that does data u is as same as data v.

        :param v: reference data
        :param u: tested data
        :return: False if validation fail.
        """
        for f in self.fields:
            fv = getattr(v, f.field_name, None)

            if isinstance(u, dict):
                fu = u.get(f.field_name, None)
            else:
                fu = getattr(u, f.field_name, None)

            if not f.validate(fv, fu):
                fprint(f'{self.cls_name}.{f.field_name} validate fail: {fv} != {fu}')
                return False

        return True

    def filename(self, data: Optional[T], **kwargs) -> str:
        """build filename for persistence instance.

        :param data: persistence instance.
        :param kwargs: overwrote keywords.
        :return: filename.
        :raise RuntimeError: filename field missing.
        :raise TypeError: wrong field.filename type.
        """
        # legacy name
        ret = ['cache', self.cls_name]

        for f in self.fields:
            if f.filename:
                if (field_value := kwargs.get(f.field_name, missing)) is missing:
                    if data is not None:
                        field_value = getattr(data, f.field_name, None)
                    else:
                        raise RuntimeError(f'field {f.field_name} is required for filename')

                if field_value is missing:
                    if f.autoinc:
                        s = '{}'
                    else:
                        s = '*'
                elif f.filename is True:
                    s = str(field_value)
                elif callable(f.filename):
                    s = str(f.filename(field_value))
                else:
                    raise TypeError()

                ret.append(f.filename_prefix + s)

        return self.filename_field_splitter.join(ret)


def _persistence_class_init(pc: PersistentClass):
    init_fields = ['self'] + [f.field_name for f in pc.fields if f.init]
    code = []
    for name in init_fields[1:]:
        code.append(f'self.{name} = {name}')
    if (af := pc.autoinc_field()) is not None:
        code.append(f'self.{af.field_name} = missing')
    return create_fn('__init__', init_fields, '\n'.join(code),
                     locals=dict(missing=missing))


def _persistence_class_str(pc: PersistentClass):
    return create_fn('__str__', (['self'], str), 'return filename(self)', locals={'filename': filename})


def _persistence_class_repr(pc: PersistentClass):
    init_fields = [f.field_name for f in pc.fields if f.init]

    code = [f'return ("{pc.cls_name}' + '{"']
    for i, name in enumerate(init_fields):
        comma = '", " ' if i > 0 else ''
        code.append(f'{comma}f"{name}={{self.{name}}}"')
    code.append('"}")')

    return create_fn('__repr__', (['self'], str), ' '.join(code))


def _persistence_class_replace(pc: PersistentClass):
    init_fields = [f.field_name for f in pc.fields if f.init]
    data_fields = [f.field_name for f in pc.fields if not f.init and not f.autoinc]
    code = [f'ret = {pc.cls_name}(']
    for name in init_fields:
        code.append(f'{name} if {name} is not missing else self.{name},')
    code.append(')')
    for name in data_fields:
        code.append(f'ret.{name} = {name} if {name} is not missing else self.{name}')
    code.append('return ret')

    return create_fn('_replace',
                     (['self', '*'] + [(name, None, 'missing') for name in (init_fields + data_fields)], pc.cls_name),
                     '\n'.join(code),
                     locals={pc.cls_name: pc.persistence_cls, 'missing': missing})


def auto_generated_content():
    """It is used to mark the function which its function body is auto generated.

    :return: nothing
    """
    raise RuntimeError('It is auto generated content')


def ensure_persistence_class(data: Union[T, type[T]]) -> PersistentClass[T]:
    """ensure **data** is a persistence class.

    :param data: instance or type
    :return: persistence info
    :raise RuntimeError: not a persistence class
    """
    if not isinstance(data, type):
        data = type(data)

    try:
        cls_info: PersistentClass = data._ast_persistence_cls_info_
    except AttributeError as e:
        raise RuntimeError(f'not a persistence_class : {data.__name__}') from e

    return cls_info


def as_dict(data: T) -> dict[str, Any]:
    """transform persistence *data* into dictionary, which field as key.

    :param data: persistence instance
    :return: dict
    """
    if data is None:
        raise TypeError('data is None')

    info = ensure_persistence_class(data)

    ret = {}
    for field in info.fields:
        try:
            ret[field.field_name] = getattr(data, field.field_name)
        except AttributeError:
            pass
    return ret


def from_dict(data_cls: type[T], d: dict[str, Any]) -> T:
    """transform dictionary

    :param data_cls:
    :param d:
    :return:
    """
    info = ensure_persistence_class(data_cls)

    def get_or_raise(key):
        try:
            return d[key]
        except KeyError:
            pass

        raise KeyError(f'missing required field : {key}')

    init = {
        f.field_name: get_or_raise(f.field_name)
        for f in info.fields
        if f.init
    }

    try:
        ret = data_cls(**init)
    except TypeError:
        if hasattr(data_cls, 'from_dict'):
            return _from_dict_factory(data_cls, info, d)
        else:
            raise
    else:
        return _from_dict_builtin(ret, info, d)


def _from_dict_builtin(ret: T, info: PersistentClass, d: dict[str, Any]) -> T:
    for f in info.fields:
        if not f.init:
            try:
                v = d[f.field_name]
            except KeyError:
                pass
            else:
                setattr(ret, f.field_name, v)
    return ret


def _from_dict_factory(data_cls: type[T], info: PersistentClass, d: dict[str, Any]) -> T:
    kwargs = {}
    for f in info.fields:
        try:
            v = d[f.field_name]
        except KeyError:
            pass
        else:
            kwargs[f.field_name] = v

    ret = data_cls.from_dict(kwargs)
    if not isinstance(ret, data_cls):
        raise TypeError()
    return ret


def save(data: T, path: P) -> None:
    """save persistence **data** under **path**.

    :param data: persistence instance
    :param path: filepath
    """
    PickleHandler(type(data), path.parent).save_persistence(data, path)


def load(data_cls: type[T], path: P) -> T:
    """Load data as **data_cls** from **path**.

    :param data_cls: data class type
    :param path: filepath
    :return: persistence instance
    """
    return PickleHandler(data_cls, path.parent).load_persistence(path)


def load_by(data_cls: type[T], path: P, **kwargs) -> T:
    """Load **data_cls** from directory **path** with fields **kwargs**.

    :param data_cls: data class type
    :param path: directory
    :param kwargs: data fields
    :return: persistence instance
    """
    handler = PickleHandler(data_cls, path.parent)
    return handler.load_persistence(handler.filepath(None, **kwargs))


def filename(result: Union[T, type[T]], **kwargs) -> str:
    """Get data persistence filename.

    :param result:
    :param kwargs: overwrite fields.
    :return: filename
    :raise RuntimeError: *result*'s autoinc field not resolved, or filename field missing.
    :raise TypeError: wrong field.filename type.
    """
    cls_info = ensure_persistence_class(result)
    if isinstance(result, type):
        result = None

    if not cls_info.is_autoinc_field_resolved(result, **kwargs):
        af = cls_info.autoinc_field()
        raise RuntimeError(f'cannot generate filepath without autoinc field {af.field_name} keywords')

    name = cls_info.filename(result, **kwargs)

    return name


class PersistenceHandler(Generic[T], metaclass=abc.ABCMeta):
    """The handler for loading and saving persistence instance."""

    @property
    @abc.abstractmethod
    def persistence_class(self) -> type[T]:
        pass

    @property
    def persistence_info(self) -> PersistentClass[T]:
        """information for persistence class"""
        return ensure_persistence_class(self.persistence_class)

    @property
    @abc.abstractmethod
    def save_root(self) -> Path:
        """saving directory"""
        pass

    def filename(self, cache: Optional[T], **kwargs) -> str:
        """build filename for persistence instance.

        :param cache: persistence instance
        :param kwargs: overwrite field value in *result*.
        :return: file name of *result*, may contains '{}' if *result*'s autoinc field not resolved
        """
        cls_info = ensure_persistence_class(self.persistence_class)
        return cls_info.filename(cache, **kwargs)

    def filepath(self, cache: Optional[T], **kwargs) -> Path:
        """build filepath for persistence instance.

        :param cache: persistence instance
        :param kwargs: overwrite field value in *result*.
        :return: file path of *result*
        :raise RuntimeError: *result*'s autoinc field not resolved, or errors from :meth:`filename`
        """
        info = ensure_persistence_class(self.persistence_class)
        if not info.is_autoinc_field_resolved(cache, **kwargs):
            af = info.autoinc_field()
            raise AutoIncFieldNotResolvedError(cache, af,
                                               f'cannot generate filepath without autoinc field {af.field_name} keywords')

        name = self.filename(cache, **kwargs)
        return self.save_root / name

    def save_persistence(self, cache: T, path: Union[str, Path] = None) -> T:
        """save persistence *result* under **path**.

        :param cache:
        :param path: save path.
        :return: *result*. autoinc field will be resolved after saving.
        :raise AutoIncFieldNotResolvedError:
        """
        info = ensure_persistence_class(cache)

        if path is None:
            if not info.is_autoinc_field_resolved(cache):
                f = info.autoinc_field()
                found = [it for _, it in self.load_all(cache, **{f.field_name: '*'})]

                u = max([
                    value
                    for it in found
                    if isinstance(value := getattr(it, f.field_name, 0), int)
                ], default=-1) + 1

                setattr(cache, f.field_name, u)

            path = self.filepath(cache)

        else:
            if not info.is_autoinc_field_resolved(cache):
                raise AutoIncFieldNotResolvedError(cache, info.autoinc_field())

            if isinstance(path, str):
                path = Path(path)

        if path.is_dir():
            raise IsADirectoryError(str(path))

        path.parent.mkdir(parents=True, exist_ok=True)
        self._save_persistence(cache, path)
        return cache

    @abc.abstractmethod
    def _save_persistence(self, cache: T, path: Path):
        pass

    def load_persistence(self, path: Union[Path, T, dict[str, Any]]) -> T:
        """Load data as **data_cls** from **path**.

        :param path: load from path.
        :return: persistence instance
        :raise IsADirectoryError:
        """
        data_cls = self.persistence_class
        ensure_persistence_class(data_cls)

        if isinstance(path, data_cls):
            path = self.filepath(path)
        elif isinstance(path, dict):
            path = self.filepath(None, **path)
        elif isinstance(path, str):
            path = Path(path)

        if path.is_dir():
            raise IsADirectoryError(str(path))

        return self._load_persistence(path)

    @abc.abstractmethod
    def _load_persistence(self, path: Path) -> T:
        pass

    def load_all(self, cache: Optional[T], **kwargs) -> Iterable[tuple[Path, T]]:
        for file in self.save_root.glob(self.filename(cache, **kwargs)):
            yield file, self.load_persistence(file)


class PickleHandler(PersistenceHandler[T]):
    """

    Support field type: all python objects.

    """

    def __init__(self, data_cls: type[T], save_root: Path, ext: str = '.pkl'):
        ensure_persistence_class(data_cls)
        self._save_path = save_root
        self._data_cls = data_cls
        self._ext = ext

    @property
    def persistence_class(self) -> type[T]:
        return self._data_cls

    @property
    def save_root(self) -> Path:
        return self._save_path

    def filename(self, result: Optional[T], **kwargs) -> str:
        return super().filename(result, **kwargs) + self._ext

    def _save_persistence(self, result: T, path: Path):
        import pickle

        with path.open('wb') as file:
            pickle.dump(as_dict(result), file)

    def _load_persistence(self, path: Path) -> T:
        import pickle

        with path.open('rb') as file:
            ret = pickle.load(file)

        data_cls = self.persistence_class
        if isinstance(ret, dict):
            return from_dict(data_cls, ret)
        elif isinstance(ret, data_cls):  # for old persistent pickle file
            return ret
        else:
            raise TypeError(f'not a {data_cls.__name__} for cache {path} : {ret}')


class GzipHandler(PersistenceHandler[T]):
    """

    Support field type: all python objects.

    """

    def __init__(self, data_cls: type[T], save_root: Path, ext: str = '.pkl.gz',
                 compression: int = 9):
        ensure_persistence_class(data_cls)
        self._save_path = save_root
        self._data_cls = data_cls
        self._ext = ext
        self._cmp = compression

    @property
    def persistence_class(self) -> type[T]:
        return self._data_cls

    @property
    def save_root(self) -> Path:
        return self._save_path

    def filename(self, result: Optional[T], **kwargs) -> str:
        return super().filename(result, **kwargs) + self._ext

    def _save_persistence(self, result: T, path: Path):
        import pickle
        import gzip

        with gzip.open(path, 'wb', compresslevel=self._cmp) as file:
            pickle.dump(as_dict(result), file)

    def _load_persistence(self, path: Path) -> T:
        import pickle
        import gzip

        with gzip.open(path, 'rb') as file:
            ret = pickle.load(file)

        return from_dict(self.persistence_class, ret)


# ========================= #
# User-Specific for 2P data #
# ========================= #

class ETLConcatable(Generic[T], metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def concat_etl(self, data: list[T]) -> T:
        pass


class PersistenceConcatError(Exception):
    pass


def validate_concat_etl_persistence(data: list[T],
                                    field_check: Optional[tuple[str, ...]] = None) -> None:
    """

    :param data:
    :param field_check: field name under persistence cls for checking
    :return:
    """
    #
    if not hasattr(data[0], 'exp_date') or not hasattr(data[0], 'animal'):
        raise RuntimeError('different dataset is not concatable')

    #
    for it in data:

        if not hasattr(it, 'plane_index'):
            raise AttributeError('not plane index field')

        if not isinstance(it.plane_index, int):
            raise TypeError('')

    if len(set([it.exp_date for it in data])) != 1:
        print(set([it.exp_date for it in data]))
        raise PersistenceConcatError('different exp date')

    if len(set([it.animal for it in data])) != 1:
        raise PersistenceConcatError('different animal')

    #
    if field_check is not None:
        for f in field_check:
            init = getattr(data[0], f)

            for it in data:
                check = getattr(it, f)

                if not is_iterable(init):
                    if check != init:
                        raise PersistenceConcatError(f' field:{f} not consistent')

                elif is_iterable(init) and isinstance(init, np.ndarray):
                    if not np.array_equal(init, check):
                        raise PersistenceConcatError(f' field:{f} not consistent')

                else:
                    for i, v in enumerate(init):
                        check_field = check
                        if check_field[i] != v:
                            raise PersistenceConcatError(f' field:{f} not consistent')



from neuralib.atlas.brainrender.core import BrainReconstructor

if __name__ == '__main__':
    brender = BrainReconstructor()
    brender.main((['--source=allen_mouse_25um', '--region=VISal,VISam,VISl,VISli,VISp,VISpl,VISpm,VISpor'], []))



from __future__ import annotations

import abc
from pathlib import Path
from typing import Any, overload

import numpy as np
import polars as pl
from polars import ColumnNotFoundError

from neuralib.stimpy import STIMPY_SOURCE_VERSION
from neuralib.util.util_verbose import fprint, printdf

__all__ = ['AbstractStimProtocol']


class AbstractStimProtocol(metaclass=abc.ABCMeta):
    """ABC class for the stimpy protocol file (.prot)

    `Dimension parameters`:

        N = numbers of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = Number of Stim Type

        C = Numbers of Cycle
    """

    name: str
    """protocol name. related to filename"""

    options: dict[str, Any]
    """protocol options"""

    visual_stimuli_dataframe: pl.DataFrame
    """visual stimuli dataframe. row number: S"""

    version: STIMPY_SOURCE_VERSION
    """date of major changes"""

    def __repr__(self):
        ret = list()

        ret.append('# general parameters')
        for k, v in self.options.items():
            ret.append(f'{k} = {v}')
        ret.append('# stimulus conditions')
        ret.append('\t'.join(self.stim_params))

        ret.append(printdf(self.visual_stimuli_dataframe, do_print=False))

        return '\n'.join(ret)

    def __init__(self, name: str,
                 options: dict[str, Any],
                 visual_stimuli: pl.DataFrame,
                 version: STIMPY_SOURCE_VERSION):

        self.name = name
        self.options = options
        self.visual_stimuli_dataframe = visual_stimuli
        self.version = version

        self._visual_stimuli = None

    @property
    def n_stimuli(self) -> int:
        """number of stimuli (S)"""
        return self.visual_stimuli_dataframe.shape[0]

    @property
    def stim_params(self) -> tuple[str, ...]:
        return tuple(self.visual_stimuli_dataframe.columns)

    # @property
    # def visual_stimuli(self) -> list[GenericStim]:  # TODO check
    #     if self._visual_stimuli is None:
    #         from rscvp.ztimpy.stim.stimulus import VisualStimulus
    #         self._visual_stimuli = VisualStimulus.from_protocol(self)
    #     return self._visual_stimuli

    @overload
    def __getitem__(self, item: int) -> pl.DataFrame:
        """get row of visual stimuli"""
        pass

    @overload
    def __getitem__(self, item: str) -> np.ndarray:
        """get header of stimuli"""
        pass

    def __getitem__(self, item: str | int) -> np.ndarray | pl.DataFrame:
        """Get protocol value of parameter *item*

        :param item: parameter name. either the row number(int) or field name (str)
        :return: protocol value
        :raises TypeError: `item` not a `str` or `int`
        """
        if isinstance(item, str):
            try:
                return self.visual_stimuli_dataframe.get_column(item).to_numpy()
            except ColumnNotFoundError as e:
                fprint(f'INVALID: {item}, select from {tuple(self.visual_stimuli_dataframe.columns)}', vtype='error')
                raise e

        elif isinstance(item, int):
            ret = {
                h: self.visual_stimuli_dataframe.row(item)[i]
                for i, h in enumerate(self.visual_stimuli_dataframe.columns)
            }
            return pl.DataFrame(ret)
        else:
            raise TypeError(f'{type(item)}')

    @classmethod
    @abc.abstractmethod
    def load(cls, file: Path | str) -> AbstractStimProtocol:
        """Load *.prot file

        :param file: file path
        """
        pass

    @property
    @abc.abstractmethod
    def is_shuffle(self) -> bool:
        """if shuffle stimulation"""
        pass

    @property
    @abc.abstractmethod
    def background(self) -> float:
        """background for non-stimulation epoch.
        Note the default value need to check in user-specific stimpy version
        """
        pass

    @property
    @abc.abstractmethod
    def start_blank_duration(self) -> int:
        """blank duration before starting the visual stimulation epoch (in sec)"""
        pass

    @property
    @abc.abstractmethod
    def blank_duration(self) -> int:
        """blank duration between each visual stimulus (in sec)"""
        pass

    @property
    @abc.abstractmethod
    def trial_blank_duration(self) -> int:
        """blank duration between trials (in sec)
        TODO check stimpy source code"""
        pass

    @property
    @abc.abstractmethod
    def end_blank_duration(self) -> int:
        """blank duration after the visual stimulation epoch"""
        pass

    @property
    @abc.abstractmethod
    def trial_duration(self) -> int:
        """trial duration"""
        pass

    @property
    def visual_duration(self) -> int:
        """total visual duration"""
        return self.trial_duration * self.n_trials

    @property
    def total_duration(self) -> int:
        """total protocol duration"""
        return self.start_blank_duration + self.visual_duration + self.end_blank_duration

    @property
    def stimulus_type(self) -> str:
        """stimulus type"""
        return self.options['stimulusType']

    @property
    def n_trials(self) -> int:
        """(T,)"""
        return self.options['nTrials']

    def to_dict(self) -> dict[str, Any]:
        """to stimpy runtime readable dict TODO"""
        pass



import functools
import os

import psutil

from neuralib.util.util_verbose import fprint

__all__ = ['profile_test']


class profile_test:
    def __init__(self, enable=False, output_file='profile.png'):
        self._enable = enable
        self._profile = None
        self._output_file = output_file

    def __enter__(self):
        if self._enable:
            import cProfile

            self._profile = cProfile.Profile()
            self._profile.enable()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._enable:
            self._profile.disable()

            if self._output_file.endswith('.dat'):
                self._profile.dump_stats(self._output_file)

            elif self._output_file.endswith('.txt'):
                import pstats
                stat = pstats.Stats(self._profile)
                with open(self._output_file, 'w') as f:
                    stat.stream = f
                    stat.print_stats()

            elif self._output_file.endswith('.png'):
                import subprocess

                f1 = self._output_file.replace('.png', '.dat')

                self._profile.dump_stats(f1)

                cmd_line = f'python -m gprof2dot -f pstats {f1} | dot -T png -o {self._output_file}'

                with subprocess.Popen(['bash', '-c', cmd_line]) as proc:
                    proc.wait()


def process_memory() -> int:
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    return mem_info.rss


def mem_profile():
    """TODO compare with memory_profiler.profile decorator"""

    def _decorator(f):
        @functools.wraps(f)
        def _wrapper(*args, **kwargs):
            before = process_memory()
            f(*args, **kwargs)
            after = process_memory()
            fprint(f'func: <{f.__name__}> consumed memory {after - before}')

        return _wrapper

    return _decorator



from __future__ import annotations

from pathlib import Path

import numpy as np
import pandas as pd
import polars as pl
from rich.pretty import pprint

from neuralib.util.util_type import PathLike


class PickleParquetViewer:

    def __init__(self, file: PathLike,
                 view_array_shape: bool = False,
                 view_iterable_len: bool = False):
        """
        Simple CLI Viewer for customized pickle/parquet file extension

        :param file: file path
        :param view_array_shape: if field as an array type, only view its shape
        :param view_iterable_len: if field as an iterable object (dict, set, list, tuple), only view its length
        """
        self.file = file
        self.suffix = Path(file).suffix

        self.view_shape_only = view_array_shape
        self.view_iterable_len = view_iterable_len

    def view(self):
        if self.suffix in ('.pkl', '.pickle'):
            self._view_pkl()
        elif self.suffix == '.parquet':
            self._view_parquet()
        else:
            raise ValueError(f'{self.suffix} not support')

    def _view_pkl(self) -> None:
        import pickle
        with open(self.file, 'rb') as f:
            data = pickle.load(f)

        if isinstance(data, dict):
            if not self.view_shape_only and not self.view_iterable_len:
                pprint(data)
            else:
                for key, val in data.items():
                    if hasattr(val, 'shape') and self.view_shape_only:
                        info = val.shape
                    elif hasattr(val, '__len__') and self.view_iterable_len:
                        info = len(val)
                    else:
                        info = val
                    pprint(f'{key}: {info}')

        elif isinstance(data, pd.DataFrame):
            data = pd.read_pickle(self.file)
            if self.view_shape_only:
                data = data.applymap(lambda x: x.shape if isinstance(x, np.ndarray) else x)

            if self.view_iterable_len:
                data = data.applymap(lambda x: len(x) if isinstance(x, (list, tuple, set, dict)) else x)

            pprint(data)

        else:
            raise TypeError(f'not supported {type(data)} pickle file')

    def _view_parquet(self):
        from neuralib.util.util_verbose import printdf
        data = pl.read_parquet(self.file)

        # only consider list[Any] column type in polars dataframe
        if self.view_iterable_len:
            for s in data.iter_columns():
                if isinstance(s.dtype, pl.List):
                    data = data.with_columns(pl.col(s.name).list.len())

        printdf(data)


def main():
    import argparse
    ap = argparse.ArgumentParser(description='Simple CLI Viewer for customized pickle/parquet file extension')

    ap.add_argument(metavar='FILE', help='pickle/parquet file path', dest='file')
    ap.add_argument('-S', '--shape', action='store_true', help='if field as an array type, only view its shape')
    ap.add_argument('-L', '--len', action='store_true',
                    help='if field as an iterable object (dict, set, list, tuple), only view its length')

    opt = ap.parse_args()

    PickleParquetViewer(opt.file, view_array_shape=opt.shape, view_iterable_len=opt.len).view()


if __name__ == '__main__':
    main()



from __future__ import annotations

import abc
from pathlib import Path
from typing import Literal, TypeVar, Generic, cast, TypedDict

import numpy as np
import polars as pl

from neuralib.util.util_type import PathLike
from .event import RigEvent, CamEvent
from .session import Session, SessionInfo
from .stimulus import StimPattern

__all__ = [
    'STIMPY_SOURCE_VERSION',
    'LOG_SUFFIX',
    'CAMERA_TYPE',
    #
    'Baselog',
    'StimlogBase',
    'AbstractStimTimeProfile'
]

STIMPY_SOURCE_VERSION = Literal['pyvstim', 'stimpy-bit', 'stimpy-git', 'debug']
LOG_SUFFIX = Literal['.log', '.riglog']
CAMERA_TYPE = Literal['facecam', 'eyecam', '1P_cam']

#
R = TypeVar('R', bound='Baselog')  # Riglog-Like
S = TypeVar('S', bound='StimlogBase')  # Stimlog-Like
P = TypeVar('P', bound='AbstractStimProtocol')  # protocol file


# =========== #
# Baselog ABC #
# =========== #

class RigConfig(TypedDict, total=False):
    source_version: STIMPY_SOURCE_VERSION
    """stimpy source version {'pyvstim', 'stimpy-bit', 'stimpy-git'}"""
    version: float
    """acquisition flag. i.e., 0.3"""
    commit_hash: str
    """git commit hash"""
    codes: dict[str, int]
    """<EVENT_TYPES>:<NUMBER>"""
    fields: tuple[str, ...]
    """column repr for the logging"""


class Baselog(Generic[S, P], metaclass=abc.ABCMeta):
    """ABC class for different stimpy/pyvstim log files. i.e., .log, .riglog"""

    log_config: RigConfig

    def __init__(self,
                 root_path: PathLike,
                 log_suffix: LOG_SUFFIX,
                 diode_offset: bool = True):
        """
        :param root_path: log file path or log directory
        :param log_suffix: log file suffix
        :param diode_offset: whether do the diode offset
        """

        if not isinstance(root_path, Path):
            root_path = Path(root_path)

        if root_path.is_dir():
            self.riglog_file = self._find_logfile(root_path, log_suffix)
        else:
            self.riglog_file = root_path

        self.log_config = self._get_log_config()
        self.version = self.log_config['source_version']

        self.dat = self._cache_asarray(self.riglog_file)

        #
        self._diode_offset = diode_offset

    @classmethod
    def _find_logfile(cls,
                      root: Path,
                      log_suffix: LOG_SUFFIX) -> Path:

        f = list(root.glob(f'*{log_suffix}'))
        if len(f) == 1:
            return f[0]

        elif len(f) == 0:
            print(f'no riglog under {root}, try to find in the subfolder...')
            for s in root.iterdir():
                if s.is_dir() and s.name.startswith('run0'):
                    try:
                        return cls._find_logfile(s, log_suffix)
                    except FileNotFoundError:
                        pass

            raise FileNotFoundError(f'no riglog file {log_suffix} under {root}')

        else:
            raise FileNotFoundError(f'more than one riglog files under {root}')

    # noinspection PyTypedDict
    def _get_log_config(self) -> RigConfig:
        """get config dict and source version for different logs from the # headers"""
        ret = RigConfig()
        with open(self.riglog_file) as f:
            for line in f:
                if '#' in line:
                    if 'RIG VERSION' in line:
                        ret['version'] = float(line.split(': ')[-1])

                    elif 'RIG GIT COMMIT HASH' in line:
                        ret['commit_hash'] = line.split(': ')[-1].strip()

                    elif 'CODES' in line:
                        codes = {}
                        content = line.replace('# CODES: ', '').strip()
                        iter_codes = content.split(',')
                        for pair in iter_codes:
                            code, num = pair.split('=')
                            code = code.strip()
                            value = int(num.strip())
                            codes[code.lower()] = value

                        ret['codes'] = codes

                    elif 'RIG CSV' in line:
                        content = line.replace('# RIG CSV: ', '').strip()
                        ret['fields'] = tuple(content.split(','))

        # infer
        if 'opto' not in ret['codes']:
            ret['source_version'] = 'stimpy-bit'

        if 'version' not in ret:
            ret['source_version'] = 'stimpy-git'

        if 'opto' in ret['codes'] and ret['codes']['opto'] == 15:
            ret['source_version'] = 'pyvstim'

        return ret

    @classmethod
    @abc.abstractmethod
    def _cache_asarray(cls, filepath: Path) -> np.ndarray:
        pass

    @abc.abstractmethod
    def get_prot_file(self) -> P:
        """
        get protocol (TypeVar ``P``)

        :return: :class:`~neuralib.stimpy.baseprot.AbstractStimProtocol()`
        """
        pass

    # ============ #
    # RIGLOG EVENT #
    # ============ #

    def _event(self, code: int) -> np.ndarray:
        """shape(sample, 2) with time and value"""
        x = self.dat[:, 0] == code
        t = self.dat[x, 2].copy() / 1000  # s
        v = self.dat[x, 3].copy()
        return np.vstack([t, v]).T

    @property
    def exp_start_time(self) -> float:
        """experimental start time (in sec)"""
        return self.dat[0, 2].copy() / 1000

    @property
    def exp_end_time(self) -> float:
        """experimental end time (in sec)"""
        return self.dat[-1, 2].copy() / 1000

    @property
    def total_duration(self) -> float:
        """experimental duration (in sec)"""
        return self.exp_end_time - self.exp_start_time

    @property
    def screen_event(self) -> RigEvent:
        """screen rig event. i.e., diode pulse"""
        return RigEvent('screen', self._event(0))

    @property
    def imaging_event(self) -> RigEvent:
        """imaging rig event. i.e., 2photon pulse"""
        return RigEvent('imaging', self._event(1))

    @property
    def position_event(self) -> RigEvent:
        """position rig event. i.e., encoder pulse"""
        return RigEvent('position', self._event(2))

    @property
    def lick_event(self) -> RigEvent:
        """lick rig event. i.e., lick meter pulse"""
        return RigEvent('lick', self._event(3))

    @property
    def reward_event(self) -> RigEvent:
        """reward rig event. i.e., reward given pulse from lick meter"""
        if self.version == 'stimpy-git':
            return RigEvent('reward', self._event(5))
        else:
            return RigEvent('reward', self._event(4))

    @property
    def lap_event(self) -> RigEvent:
        """lap rig event. i.e., optic sensing for the reflective taps"""
        if self.version == 'stimpy-git':
            return RigEvent('lap', self._event(6))
        else:
            return RigEvent('lap', self._event(5))

    @property
    def act_event(self) -> RigEvent:
        """todo"""
        raise NotImplementedError('')

    @property
    def opto_event(self) -> RigEvent:
        """todo"""
        raise NotImplementedError('')

    class CameraEvent:
        """camera event"""
        camera: dict[CAMERA_TYPE, int]

        def __init__(self, rig: R):
            """
            :param rig:``Baselog``
            """
            self.rig = rig

            if rig.version == 'stimpy-git':
                self.camera = {
                    'facecam': 7,
                    'eyecam': 8,
                    '1P_cam': 9,
                }
            else:
                self.camera = {
                    'facecam': 6,
                    'eyecam': 7,
                    '1P_cam': 8,
                }

        def __getitem__(self, item: CAMERA_TYPE) -> CamEvent:
            if item not in self.camera:
                raise KeyError('cam id not found')
            return CamEvent(item, self.rig._event(self.camera[item]))

    @property
    def camera_event(self) -> CameraEvent:
        """camera event. including {'facecam', 'eyecam', '1P_cam'} implemented by __getitem__()"""
        return self.CameraEvent(self)

    @property
    def stim_prot_file(self) -> Path:
        """protocol file path"""
        return self.riglog_file.with_suffix('.prot')

    @abc.abstractmethod
    def stimlog_data(self) -> S:
        """get stimlog (TypeVar ``S``)

        :return: :class:`~neuralib.stimpy.baselog.StimlogBase()`
        """
        pass


# =========== #
# Stimlog ABC #
# =========== #

class StimlogBase(Generic[R], metaclass=abc.ABCMeta):
    """ABC for stimulation logging. i.e., .log or .stimlog

    `Dimension parameters`:

        N = number of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = number of Stim Type

        P = number of acquisition sample pulse

    """

    def __init__(self,
                 riglog: R,
                 file_path: PathLike | None):
        """
        :param riglog: :class:`Baselog`
        :param file_path: filepath of stimlog. could be None if shared log (pyvstim case)
        """
        self.riglog_data = riglog
        if file_path is not None:
            self.stimlog_file = Path(file_path)

    @abc.abstractmethod
    def _reset(self) -> None:
        """used for assign attributes"""
        pass

    @property
    @abc.abstractmethod
    def exp_start_time(self) -> float:
        """experimental start time (in sec, synced to riglog time with diode offset)"""
        pass

    @property
    @abc.abstractmethod
    def exp_end_time(self) -> float:
        """experimental end time (in sec, synced to riglog time with diode offset)"""
        pass

    @property
    def stim_start_time(self) -> float:
        """the first stimulation start time (in sec, synced to riglog time with diode offset)"""
        return cast(float, self.stimulus_segment[0, 0])

    @property
    def stim_end_time(self) -> float:
        """the last stimulation end time (in sec, synced to riglog time with diode offset)"""
        return cast(float, self.stimulus_segment[-1, 1])

    @property
    @abc.abstractmethod
    def stimulus_segment(self) -> np.ndarray:
        """stimulation time segment (on-off) in sec (N, 2)"""
        pass

    @abc.abstractmethod
    def session_trials(self) -> dict[Session, SessionInfo]:
        """get the session:SessionInfo dictionary (experimental and user-specific)"""
        pass

    @property
    @abc.abstractmethod
    def time_offset(self) -> float:
        """time (in sec) to sync stimlog time to riglog"""
        pass

    @abc.abstractmethod
    def get_stim_pattern(self) -> StimPattern:
        """get pattern foreach stimulation"""
        pass

    @abc.abstractmethod
    def get_time_profile(self) -> AbstractStimTimeProfile:
        """get time profile"""
        pass


# ================= #
# Stim Time Profile #
# ================= #

class AbstractStimTimeProfile(Generic[S], metaclass=abc.ABCMeta):
    """
    ABC for stimulation time profile

    `Dimension parameters`:

        N = number of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = number of Stim Type

        C = number of Cycle

    """
    stim: S

    @property
    @abc.abstractmethod
    def unique_stimuli_set(self) -> pl.DataFrame:
        """
        rows: number equal to N

        fields: i_stim (index of stim type); i_trials (repetitive trials in the given stim type)
        """
        pass

    @property
    @abc.abstractmethod
    def n_trials(self) -> int:
        """T"""
        pass

    @abc.abstractmethod
    def get_time_interval(self) -> np.ndarray:
        """(N, 2) with (start, end). suppose the same as :attr:`neuralib.stimpy.baselog.StimlogBase.stimulus_segment`"""
        pass



"""
Stimpy parser
==============

:author:
    Yu-Ting Wei


This module provide parser utilities for data acquired from stimpy (lab internal usage)


Including
----------

- logs (.log, .riglog, .stimlog, .camlog)
- protocol file (.prot)
- preference file (.prefs)
- reproduce the experimental stimuli (under dev)
- other utilities (synced time, ...)


**Design Only For the branch follows the git history**
--------------------------------------------------------

- pyvstim system master branch (legacy acquisition version)
- stimpy bitbucket master branch
- stimpy github master branch


Example of riglog data
-----------------------------

.. code-block:: python

    from neuralib.stimpy import RiglogData, PyVlog

    root_path = ...  # directory with all the stimpy output
    rig = RiglogData(root_path)  # stimpy
    # rig = PyVlog(root_path)  # if pyvstim

    print(rig.lap_event.time)  # get lick event time
    print(rig.camera_event['facecam'].time)  # get face camera event time




Example of stimlog data
-----------------------------

.. code-block:: python

    stimlog = rig.stimlog_data()

    # stimulation time segment (on-off) in sec (N, 2), already synced time to riglog
    print(stimlog.stimulus_segment)

    # get foreach stimulus generator (index, stimulus_time, sf, tf, ori)
    pattern = stimlog.get_stim_pattern()
    for si, st, sf, tf, dire in pattern.foreach_stimulus():
        print(f'idx:{si}, time:{st}, sf:{sf}, tf:{tf}, dire:{dire}')




Example of protocol
-----------------------------

.. code-block:: python

    prot = rig.get_prot_file()
    print(prot)  # see the protocol content
    print(prot.visual_stimuli_dataframe)  # see the parsed protocol dataframe

"""

from .baselog import *
from .baseprot import *
from .camlog import *
from .event import *
from .session import *
from .stimpy_core import *
from .stimpy_git import *
from .stimpy_pyv import *
from .stimulus import *
from .util import *



"""
User-specific experimental session
====================================

- Aim to know the time information about of a specific session in a given protocol


Pipeline
---------------

- Use protocol name to infer which type of protocol ``ProtocolAlias``
- Base on riglog/stimlog information to create a dictionary with ``Session``: ``SessionInfo`` pairs for a specific protocol
- Use the methods in ``SessionInfo`` to do the masking/slicing... with the ``RigEvent``

"""

from __future__ import annotations

from pathlib import Path
from typing import NamedTuple, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .baselog import StimlogBase

import numpy as np
from typing_extensions import TypeAlias

__all__ = [
    'ProtocolAlias',
    'get_protocol_name',
    'get_protocol_sessions',
    #
    'Session',
    'SessionInfo'
]

ProtocolAlias: TypeAlias = str
Session: TypeAlias = str


def get_protocol_name(filename: Path | 'StimlogBase') -> ProtocolAlias:
    from .baselog import StimlogBase
    if isinstance(filename, StimlogBase):
        filename = filename.stimlog_file

    filename = filename.stem
    if '_ori_sqr_12dir_2tf_3sf_bas' in filename or 'ori' in filename:
        return 'visual_open_loop'
    elif '_LDL' in filename:
        return 'light_dark_light'
    elif '_black' in filename:
        return 'grey'
    else:
        raise RuntimeError(f'unknown fname >> {filename}')


def get_protocol_sessions(stim: 'StimlogBase') -> list[SessionInfo]:
    protocol = get_protocol_name(stim)
    if protocol == 'visual_open_loop':
        return _get_protocol_sessions_vol(stim)
    elif protocol == 'light_dark_light':
        return _get_protocol_sessions_ldl(stim)
    elif protocol == 'grey':
        return _get_protocol_sessions_grey(stim)
    else:
        raise RuntimeError(f'unknown protocol >> {protocol}')


def _get_protocol_sessions_vol(stim: 'StimlogBase') -> list[SessionInfo]:
    """get session info for visual open loop protocol"""
    t0 = stim.riglog_data.exp_start_time
    t1 = stim.stim_start_time  # diode synced
    t2 = stim.stim_end_time  # diode synced
    t3 = stim.riglog_data.exp_end_time
    return [
        SessionInfo('light', (t0, t1)),
        SessionInfo('visual', (t1, t2)),
        SessionInfo('dark', (t2, t3)),
        SessionInfo('all', (t0, t3))
    ]


def _get_protocol_sessions_ldl(stim: 'StimlogBase') -> list[SessionInfo]:
    # diode signal is no longer reliable, use .prot file value instead
    from .stimpy_core import StimpyProtocol
    prot = StimpyProtocol.load(stim.stimlog_file.with_suffix('.prot'))

    t0 = stim.riglog_data.exp_start_time
    t1 = prot.start_blank_duration
    t2 = prot.total_duration - prot.end_blank_duration
    t3 = stim.riglog_data.exp_end_time

    # print('test t', t0, t1, t2, t3)  # todo total time exceed around 4s?
    return [
        SessionInfo('light_bas', (t0, t1)),
        SessionInfo('dark', (t1, t2)),
        SessionInfo('light_end', (t2, t3)),
        SessionInfo('all', (t0, t3)),
    ]


def _get_protocol_sessions_grey(stim: 'StimlogBase') -> list[SessionInfo]:
    t0 = stim.riglog_data.exp_start_time
    t3 = stim.riglog_data.exp_end_time
    return [SessionInfo('all', (t0, t3))]


# ============ #

class SessionInfo(NamedTuple):
    """session name and the corresponding time start-end"""

    name: Session
    """name of this session"""

    time: tuple[float, float]
    """time start/end of this session"""

    def time_mask_of(self, t: np.ndarray) -> np.ndarray:
        """
        create a mask for time array *t*.

        :param t: 1d time array
        :return: mask for this session
        """
        return np.logical_and(self.time[0] < t, t < self.time[1])

    def in_range(self, time: np.ndarray,
                 value: np.ndarray = None,
                 error=True) -> tuple[Any, Any]:
        """
        Get the range (the first and last value) of value array in this session.

        :param time: 1d time array (T,)
        :param value: 1d value array. Shape should as same as *time* (T,)
        :param error: raise an error when empty.
        :return: tuple of first and last `value` or `time`.
        """
        x = self.time_mask_of(time)
        if value is not None:
            t = value[x]
        else:
            t = time[x]

        if len(t) == 0:
            if error:
                raise ValueError('empty in extracting value or time from time mask')
            return np.nan, np.nan

        return t[0], t[-1]

    def in_slice(self, time: np.ndarray,
                 value: np.ndarray,
                 error=True) -> slice:
        """
        Get the slice of value in this session

        :param time: 1d time array (T,)
        :param value: 1d value array. Shape should as same as *time* (T,)
        :param error: raise an error when empty.
        :return: slice of `value`
        """
        if not np.issubdtype(value.dtype, np.integer):
            raise ValueError()

        v = value[self.time_mask_of(time)]

        if len(v) == 0:
            if error:
                raise ValueError('empty in extracting value or time from time mask')
            return slice(0, 0)

        return slice(int(v[0]), int(v[-1]))



from __future__ import annotations

import numpy as np

__all__ = [
    'normalize_sequences',
    'handle_invalid_value'
]


def normalize_sequences(frames: list[np.ndarray] | np.ndarray,
                        handle_invalid: bool = True,
                        gamma_correction: bool = False,
                        gamma_value: float = 0.5,
                        to_8bit: bool = False) -> list[np.ndarray]:
    """
    Do the normalization for the image sequences

    :param frames: list of image array
    :param handle_invalid: handle Nan and negative value
    :param gamma_correction: to the gamma correction
    :param gamma_value: gamma correction value
    :param to_8bit: to 8bit images
    :return: list of normalized image array
    """
    if handle_invalid:
        frames = handle_invalid_value(frames)

    if gamma_correction:
        frames = [np.power(f, gamma_value) for f in frames]

    # find global min and max across all frames
    global_min = min(frame.min() for frame in frames)
    global_max = max(frame.max() for frame in frames)

    # normalize and scale to 0-255
    ret = [(frame - global_min) / (global_max - global_min) * 255 for frame in frames]

    if to_8bit:
        ret = [frame.astype('uint8') for frame in ret]

    return ret


def handle_invalid_value(frames: list[np.ndarray] | np.ndarray) -> list[np.ndarray]:
    """Handle NaN and negative values and ensure all values are >= 0"""
    frames = [np.nan_to_num(frame, nan=0, posinf=0, neginf=0) for frame in frames]
    frames = [np.clip(frame, 0, None) for frame in frames]
    return frames



from __future__ import annotations

from typing import NamedTuple, Iterable, TYPE_CHECKING

import numpy as np
from typing_extensions import TypeAlias, Self

if TYPE_CHECKING:
    from neuralib.stimpy.baselog import R

__all__ = [
    'Degree',
    'SF',
    'TF',
    'SFTF',
    'VisualParas',
    #
    'StimPattern'
]

Degree: TypeAlias = int  # degree
SF: TypeAlias = float  # cyc/deg
TF: TypeAlias = int  # Hz
SFTF: TypeAlias = tuple[SF, TF]
VisualParas: TypeAlias = tuple[SF, TF, Degree]


class StimPattern(NamedTuple):
    """Stimulus Parameters

    `Dimension parameters`:

        N = numbers of visual stimulation (on-off pairs) = (T * S)
    """
    time: np.ndarray
    """stim on-off in sec (N, 2)"""
    direction: np.ndarray  # int (degree) (N,)
    """degree (N,)"""
    sf: np.ndarray  # float (cyc/deg) (N,)
    """cyc/deg (N,)"""
    tf: np.ndarray  # int (Hz) (N,)
    """hz (N,)"""
    contrast: np.ndarray
    """stimulus contrast (N,)"""
    duration: np.ndarray
    """theoretical duration in prot file, not actual detected using diode"""

    @classmethod
    def of(cls, rig: 'R') -> Self:
        """
        init from Baselog children class

        :param rig: :class:`~neuralib.stimpy.baselog.Baselog`
        :return: :class:`StimPattern`
        """
        return rig.stimlog_data().get_stim_pattern()

    @property
    def sf_set(self) -> np.ndarray:
        """unique sf_set"""
        return np.array(sorted(self.sf_i().keys()))

    @property
    def tf_set(self) -> np.ndarray:
        """unique tf_set"""
        return np.array(sorted(self.tf_i().keys()))

    @property
    def n_sf(self) -> int:
        """number of sf set"""
        return len(self.sf_set)

    @property
    def n_tf(self) -> int:
        """number of tf set"""
        return len(self.tf_set)

    @property
    def n_sftf(self) -> int:
        """number of sftf combination"""
        return len(self.sftf_i())

    @property
    def n_dir(self) -> int:
        """number of direction"""
        return len(self.dir_i())

    def dir_i(self) -> dict[Degree, int]:
        """deg:index dict"""
        return {it: i for i, it in enumerate(sorted(np.unique(self.direction)))}

    def sf_i(self) -> dict[SF, int]:
        """sf:index dict"""
        return {it: i for i, it in enumerate(sorted(np.unique(self.sf)))}

    def tf_i(self) -> dict[TF, int]:
        """sf:index dict"""
        return {it: i for i, it in enumerate(sorted(np.unique(self.tf)))}

    # previous plot use tfsf as condition idx
    def sftf_i(self) -> dict[SFTF, int]:
        """sf, tf combination. (sf , tf):y"""
        return {
            it: i
            for i, it in enumerate([
                (sf, tf)
                for sf in sorted(np.unique(self.sf))
                for tf in sorted(np.unique(self.tf))
            ])
        }

    def sftfdir_i(self) -> dict[VisualParas, int]:
        """sf, tf, ori combination. (sf , tf , ori):y"""
        return {
            it: i  # (sf , tf, ori): index
            for i, it in enumerate([
                (sf, tf, ori * 30)
                for sf in sorted(np.unique(self.sf))
                for tf in sorted(np.unique(self.tf))
                for ori in range(12)
            ])
        }

    def foreach_stimulus(self) -> Iterable[tuple[int, np.ndarray, SF, TF, Degree]]:
        """Generator for (index, stimulus_time, sf, tf, ori)"""
        for si, st in enumerate(self.time):
            yield si, st, self.sf[si], self.tf[si], self.direction[si]

    def get_stim_time(self) -> float:
        """get approximate stim time if the same duration. i.e., for plotting purpose"""
        return np.mean(self.time[:, 1] - self.time[:, 0])



from __future__ import annotations

import abc
from pathlib import Path
from typing import Any, NewType, Literal

import numpy as np
from tifffile import tifffile
from typing_extensions import TypeAlias

from neuralib.util.util_type import PathLike
from neuralib.util.util_verbose import fprint

__all__ = [
    #
    'ZPROJ_TYPE',
    'ZEISS_CZI_CHANNELS_ORDER',
    'ZEISS_LSM_CHANNELS_ORDER',
    #
    'SceneIdx',
    'DimCode',
    #
    'AbstractConfocalScanner',
    'parse_tif_meta'
]

ZEISS_CZI_CHANNELS_ORDER = ['Greens', 'Blues', 'Reds']
ZEISS_LSM_CHANNELS_ORDER = ['Blues', 'Reds', 'Greens']

ZPROJ_TYPE = Literal['avg', 'max', 'min', 'std', 'median']

SceneIdx: TypeAlias = int  # 0-base scan position
DimCode = NewType('DIMCODE', str)


class AbstractConfocalScanner(metaclass=abc.ABCMeta):
    """ABC for the confocal image data"""

    n_scenes: int
    """positions scan"""

    meta: dict[str, Any]
    """metadata dict"""

    @classmethod
    @abc.abstractmethod
    def load(cls, filepath: PathLike):
        pass

    @property
    @abc.abstractmethod
    def width(self) -> dict[SceneIdx, int]:
        """X"""
        pass

    @property
    @abc.abstractmethod
    def height(self) -> dict[SceneIdx, int]:
        """Y"""
        pass

    @property
    @abc.abstractmethod
    def n_channels(self) -> dict[SceneIdx, int]:
        """number of fluorescence channels. C"""
        pass

    @property
    @abc.abstractmethod
    def n_zstacks(self) -> dict[SceneIdx, int]:
        """number of stacks in z axis. Z"""
        pass

    @abc.abstractmethod
    def get_dim_code(self) -> DimCode:
        """get the `DimCode`


        `Dimension parameters (DimCode)`:

            V - view

            H - phase

            I - illumination

            S - scene

            R - rotation

            T - time

            C - channel

            Z - z plane (height)

            M - mosaic tile, mosaic images only

            Y - image height

            X - image width

            A - samples, BGR/RGB images only
        """
        pass

    def get_pixel2mm_factor(self):
        pass

    def _zproj(self, stacks: np.ndarray,
               zproj_type: ZPROJ_TYPE = 'max') -> np.ndarray:
        if zproj_type == 'avg':
            img = np.mean(stacks, axis=0)
        elif zproj_type == 'median':
            img = np.median(stacks, axis=0)
        elif zproj_type == 'max':
            img = np.max(stacks, axis=0)
        else:
            raise NotImplementedError('')

        return img


# ===== #

def parse_tif_meta(file: Path, **kwargs) -> dict[str, Any]:
    """
    :return: meta dict
    """
    if file.suffix not in ('.tif', '.tiff', '.lsm'):
        raise ValueError(f'{file.suffix} not support')

    meta_collect = {}
    with tifffile.TiffFile(file, **kwargs) as tif:
        for attr in dir(tif):
            if 'meta' in attr:
                meta = getattr(tif, attr)
                if meta is not None:
                    if isinstance(meta, dict):  # i.e., lsm
                        meta_collect.update(meta)

                    elif isinstance(meta, tuple):  # tuple[dict[str, Any]]
                        for it in meta:

                            # if isinstance(it, str):
                            #     meta_collect[]

                            for k, v in it.items():
                                if k in meta_collect:
                                    raise RuntimeError(f'{k} already existed')
                            meta_collect.update(it)
                    else:
                        fprint(f'meta type {type(meta)} parse fail', vtype='warning')

    return meta_collect



from typing import Optional

import numpy as np

from neuralib.argp import argument, str_tuple_type
from neuralib.atlas.brainrender.core import BrainReconstructor, RoiType
from neuralib.atlas.util import iter_source_coordinates

__all__ = ['RoisReconstructor']

# user-specific lookup table for map the ipsi/contra hemisphere to brainrender
_HEMI_LUT = {
    'right': 'ipsi',
    'left': 'contra',
    'both': 'both'
}


class RoisReconstructor(BrainReconstructor):
    DESCRIPTION = 'For labelled rois reconstruction from 2dccf pipeline'

    only_roi_region: str = argument(
        '--roi-region',
        metavar='NAME,...',
        type=str_tuple_type,
        default=[],
        help='only show rois in region(s)'
    )

    region_col: Optional[str] = argument(
        '--region-col',
        metavar='MERGE_AC..',
        default=None,
        help='if None, auto infer, and check the lowest merge level contain all the regions specified'
    )

    def load_merge_points(self, rois: np.ndarray = None) -> RoiType:
        """
        load from 2dccf pipeline, with different channels of fluorescence labelled points

        :param rois: (N', 3) with xyz(ap, dv, ml). i.e., prelocate roi, reference points ...
        :return:
        """
        ret = []
        if rois is not None:
            if rois.shape[1] != 3:
                raise ValueError('')
            ret.append(rois)

        if self.csv_file is not None:
            iter_coords = iter_source_coordinates(self.csv_file,
                                                  only_areas=self.only_roi_region,
                                                  region_col=self.region_col,
                                                  hemisphere=_HEMI_LUT[self.hemisphere])
            for sc in iter_coords:
                ret.append(sc.coordinates)

        return ret

    def load(self):
        if self.csv_file is not None:
            rois_list = self.load_merge_points()
            self.add_points(rois_list)


if __name__ == '__main__':
    RoisReconstructor().main()



from __future__ import annotations

from pathlib import Path
from typing import Literal, Callable

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from scipy.ndimage import gaussian_filter1d

from neuralib.plot.colormap import insert_colorbar
from neuralib.util.util_type import ArrayLike
from neuralib.util.util_verbose import fprint

__all__ = [
    'plot_2d_dots',
    'plot_regression_cc',
    'plot_joint_scatter_histogram',
    'plot_histogram_cutoff'
]


# ========= #
# Dots plot #
# ========= #

def plot_2d_dots(ax: Axes,
                 x: ArrayLike,
                 y: ArrayLike,
                 size: ArrayLike, *,
                 with_color: bool = False,
                 with_legends: bool = True,
                 size_type: Literal['area', 'radius'] = 'radius',
                 size_factor: float = 3000,
                 **kwargs):
    """

    :param ax:
    :param x:
    :param y:
    :param size:
    :param with_color:
    :param with_legends:
    :param size_type: whether the value corresponding to circle radius or surface area.
    :param size_factor:
    :param kwargs:
    :return:
    """
    size = np.array([size]).flatten()
    x, y = np.meshgrid(x, y, indexing='ij')
    #
    if size_type == 'radius':
        func = lambda it: it ** 2 * size_factor
    elif size_type == 'area':
        func = lambda it: it * size_factor
    else:
        raise ValueError('')

    s = func(size)
    #
    if with_color:
        im = ax.scatter(x.ravel(), y.ravel(), s=s, c=size, cmap='viridis', clip_on=False,
                        vmin=0, vmax=1, **kwargs)
        insert_colorbar(ax, im)
    else:
        ax.scatter(x.ravel(), y.ravel(), s=s, c='k', clip_on=False, **kwargs)

    if with_legends:
        ax_size_legend(ax, size, func)


def ax_size_legend(ax: Axes,
                   value: list[int] | np.ndarray,
                   f: Callable[[float], float] = None):
    """
    add the label and legend of the size in scatter plot

    :param ax:
    :param value: values reflect to size
    :param f: amplified callable
    :return:
    """
    if f is None:
        f = lambda it: (it ** 2) * 3000

    vsize = np.linspace(0, np.max(value), num=5)

    for s in vsize:
        ax.scatter([], [], s=f(s), c='k', label=str(s))

    h, l = plt.gca().get_legend_handles_labels()
    plt.legend(h[1:], l[1:], labelspacing=1.2, title="value", borderpad=1,
               frameon=True, framealpha=0.6, edgecolor="k", facecolor="w")


# ========== #
# Regression #
# ========== #


def plot_regression_cc(ax: Axes,
                       x: np.ndarray,
                       y: np.ndarray,
                       order: int = 1,
                       show_cc: bool = True,
                       bins=10,
                       **kwargs):
    """
    Regression to see the relationship between x and y

    :param ax:
    :param x:
    :param y:
    :param order:
    :param bins:
    :param show_cc:
    :param kwargs:
    :return:
    """
    import seaborn as sns

    sns.regplot(x=x, y=y, ax=ax, order=order,
                scatter_kws={
                    'alpha': 0.5,
                    'color': 'grey',
                    's': 8,
                    'edgecolors': 'none'
                },
                line_kws={
                    'color': 'black'
                })
    _hist_line_plot(ax, x, y, bins)

    ax.set(**kwargs)
    ax.set_aspect(1.0 / ax.get_data_ratio(), adjustable='box')

    if show_cc:
        cc = calculate_cc(x, y)
        ax.set_title(f'r = {round(cc, 4)}', fontstyle='italic')


def _hist_line_plot(ax,
                    x: np.ndarray,
                    y: np.ndarray,
                    bins=20,
                    ptype: Literal['median', 'mean'] = 'median'):
    """
    average or pickup the median of the y value in certain bins
    :param ax:
    :param x: (N, )
    :param y: (N, )
    :param bins:
    :param ptype
    :return:
    """

    if ptype == 'mean':

        a, edg = np.histogram(x, weights=y, bins=bins)
        n = np.histogram(x, bins)[0]  # (B, )

        weights = np.divide(a, n, out=np.zeros_like(a, dtype=float), where=n != 0)  # avoid true divide
        weights = gaussian_filter1d(weights, 1)

        ax.plot(edg[1:], weights, 'r--', alpha=0.5)

    elif ptype == 'median':
        counts, edg = np.histogram(x, weights=y, bins=bins)

        medians = []
        for i in range(len(counts)):
            bin_data = y[(x >= edg[i]) & (x < edg[i + 1])]
            median_val = np.nanmedian(bin_data)
            medians.append(median_val)

        ax.plot(edg[1:], medians, 'r--', alpha=0.5)

    else:
        raise ValueError(f'{ptype}')


def calculate_cc(x: np.ndarray, y: np.ndarray) -> float:
    from scipy.stats import pearsonr
    try:
        cc = pearsonr(x, y)[0]
    except ValueError as e:
        print('input shape ERROR:', x.shape, y.shape)
        raise e

    return cc


# ====== #
# Others #
# ====== #


def plot_histogram_cutoff(ax: Axes,
                          value: ArrayLike,
                          cutoff: float,
                          mask: np.ndarray | None = None,
                          **kwargs) -> None:
    """
    Plot the histogram with a cutoff value

    :param ax:
    :param value: (N,) 1d array
    :param cutoff: cutoff (threshold) value for the certain value, >= represents pass
    :param mask: (N,) mask for value. i.e., cell selection
    :param kwargs: passed to `ax.set`
    """
    if value.ndim != 1:
        raise ValueError('value must be 1d array')

    if mask is not None:
        value = value[mask]

    if cutoff > np.max(value) or cutoff < np.min(value):
        fprint(f'{cutoff} should be within {np.min(value)} and {np.max(value)}', vtype='warning')

    # ax.hist(value, 50, density=True, color='grey')
    sns.histplot(value, bins=30, kde=True, ax=ax, color='grey', stat='percent', element='step')
    ax.axvline(cutoff, color='r', linestyle='--', zorder=1)
    ax.set(**kwargs)
    ax.set_aspect(1.0 / ax.get_data_ratio(), adjustable='box')


def plot_joint_scatter_histogram(x: np.ndarray,
                                 y: np.ndarray,
                                 show_cc: bool = True,
                                 output: Path | None = None,
                                 **kwargs):
    """plot the linear correlation scatter and histogram between two variables"""

    sns.set(style='white', font_scale=1.2)

    g = sns.JointGrid(x=x, y=y, height=5)
    g = g.plot_joint(sns.regplot, color='grey')
    g = g.plot_marginals(sns.histplot, kde=False, bins=12, color='grey')

    ax = g.ax_joint
    ax.set(**kwargs)

    if show_cc:
        cc = calculate_cc(x, y)
        ax.text(0.5, 0.8, f'r = {round(cc, 4)}', fontstyle='italic',
                horizontalalignment='center',
                verticalalignment='center',
                transform=ax.transAxes)

    if output is not None:
        plt.savefig(output)
    else:
        plt.show()
